"""Parse EnSpire files and optionally build titrations using a note file."""

from clophfit.prenspire.prenspire import (
    CsvLineError,
    EnspireFile,
    ManyLinesFoundError,
    Note,
    verbose_print,
)

__all__ = [
    "CsvLineError",
    "EnspireFile",
    "Note",
    "ManyLinesFoundError",
    "verbose_print",
]



"""Test cases for the binding functions module."""

import re
import warnings

import numpy as np
import pytest

from clophfit import binding
from clophfit.binding.fitting import DataArrays, Dataset, fit_binding_glob


def test_kd() -> None:
    """It returns int float np.ndarray[float]."""
    kd = binding.fitting.kd
    assert kd(10, 7, 6.0) == 11
    assert kd(10, 7, 7) == 20
    assert kd(10, 7, 8) == 110.0
    np.testing.assert_allclose(kd(10.0, 7.0, np.array([6.0, 8.0])), [11.0, 110.0])
    np.testing.assert_allclose(kd(10, 7, np.array([7, 6, 8.0])), [20, 11, 110])


class TestFitBinding:
    """Test fit_binding()."""

    def test_fit_binding_glob(self) -> None:
        """It fits pH and Cl individual titration curves."""
        # First pH titration
        x = np.array([9.0, 8.0, 7.0, 6.0, 5.0])
        y = np.array(
            [1.99, 1.909, 1.5, 1.0909, 1.0099]
        )  # This is generated by dummy_model with K=7, S0=2, S1=1
        f_res = fit_binding_glob(Dataset(x, y, is_ph=True))
        assert f_res is not None
        assert f_res.result is not None
        result = f_res.result
        assert result.success is True
        assert np.isclose(result.params["K"].value, 7, 1e-4)
        assert np.isclose(result.params["S0_default"].value, 2, 1e-4)
        assert np.isclose(result.params["S1_default"].value, 1, 1e-5)
        # pH
        x = np.array([3.0, 5, 7, 9, 11.0])
        y = np.array([1.9991, 1.991, 1.5, 1.009, 1.0009])
        f_res = fit_binding_glob(Dataset(x, y, is_ph=True))
        assert f_res is not None
        result = f_res.result
        assert result.success is True
        assert np.isclose(result.params["K"].value, 7, 1e-5)
        assert np.isclose(result.params["S0_default"].value, 1, 1e-4)
        assert np.isclose(result.params["S1_default"].value, 2, 1e-4)
        # Cl
        x = np.array([0, 5.0, 10, 40, 160, 1000])
        y = np.array([2.0, 1.33333333, 1.0, 0.4, 0.11764706, 0.01980198])
        f_res = fit_binding_glob(Dataset(np.array(x), y, is_ph=False))
        assert f_res is not None
        result = f_res.result
        assert result.success is True
        assert np.isclose(result.params["K"].value, 10, 1e-5)
        assert np.isclose(result.params["S0_default"].value, 2, 1e-4)
        assert np.isclose(result.params["S1_default"].value, 0, 1e-4)


def test_dataset_single_array_no_nan() -> None:
    """Clean nan."""
    x = np.array([1, 2, 3])
    y = np.array([4, 5, 6])
    dataset = Dataset(x, y)
    assert len(dataset) == 1
    assert isinstance(dataset["default"], DataArrays)
    assert np.array_equal(dataset["default"].x, x)
    assert np.array_equal(dataset["default"].y, y)


def test_dataset_single_array_with_nan() -> None:
    """Clean nan."""
    x = np.array([1, 2, 3, 4])
    y = np.array([5, 6, np.nan, 8])
    dataset = Dataset(x, y)
    assert len(dataset) == 1
    assert isinstance(dataset["default"], DataArrays)
    assert np.array_equal(dataset["default"].x, np.array([1, 2, 4]))
    assert np.array_equal(dataset["default"].y, np.array([5, 6, 8]))


def test_dataset_dict_no_nan() -> None:
    """Clean nan."""
    x = {"a": np.array([1, 2, 3]), "b": np.array([4, 5, 6])}
    y = {"a": np.array([7, 8, 9]), "b": np.array([10, 11, 12])}
    dataset = Dataset(x, y)
    assert len(dataset) == 2
    assert np.array_equal(dataset["a"].x, x["a"])
    assert np.array_equal(dataset["a"].y, y["a"])
    assert np.array_equal(dataset["b"].x, x["b"])
    assert np.array_equal(dataset["b"].y, y["b"])


def test_dataset_dict_with_nan() -> None:
    """Clean nan."""
    x = {"a": np.array([1, 2, 3]), "b": np.array([4, 5, 6])}
    y = {"a": np.array([7, np.nan, 9]), "b": np.array([10, 11, np.nan])}
    dataset = Dataset(x, y)
    assert len(dataset) == 2
    assert np.array_equal(dataset["a"].x, np.array([1, 3]))
    assert np.array_equal(dataset["a"].y, np.array([7, 9]))
    assert np.array_equal(dataset["b"].x, np.array([4, 5]))
    assert np.array_equal(dataset["b"].y, np.array([10, 11]))


def test_dataset_single_x_dict_y_with_nan() -> None:
    """Clean nan."""
    x = np.array([1, 2, 3, 4])
    y = {"a": np.array([5, 6, np.nan, 8]), "b": np.array([9, 10, np.nan, 12])}
    dataset = Dataset(x, y)
    assert len(dataset) == 2
    assert np.array_equal(dataset["a"].x, np.array([1, 2, 4]))
    assert np.array_equal(dataset["a"].y, np.array([5, 6, 8]))
    assert np.array_equal(dataset["b"].x, np.array([1, 2, 4]))
    assert np.array_equal(dataset["b"].y, np.array([9, 10, 12]))


def test_dataset_clean_data() -> None:
    """Clean too small datasets."""
    x = np.array([1, 2, 3])
    y = np.array([2, 3, 4])
    # Create Dataset
    ds = Dataset(x, {"y0": y}, is_ph=True)
    # Check initial keys
    assert "y0" in ds
    # Call clean_data and check for warning
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")  # Always record warnings
        ds.clean_data(4)
        assert len(w) == 1  # Check that one warning was issued
        assert str(w[0].message) == (
            "Removing key 'y0' from Dataset: number of parameters (4) exceeds "
            "number of data points (3)."
        )
    # Check keys after clean_data
    assert "y0" not in ds


def test_dataset_class() -> None:
    """Test Dataset and DataPair classes.

    This includes tests for:
    - Correct initialization with various input types (arrays and dictionaries).
    - Correct storage and retrieval of data pairs in the Dataset.
    - Appropriate error handling for mismatched keys and unequal data lengths.
    """
    # Define some test data
    x1 = np.array([5.5, 7.0, 8.5])
    x2 = np.array([6.8, 7.0, 7.2])
    y1 = np.array([2.1, 1.6, 1.1])
    y2 = np.array([1.8, 1.6, 1.4])
    # Test the case where x and y are both single ArrayF.
    # The key should default to "default".
    ds = Dataset(x1, y1)
    assert np.array_equal(ds["default"].x, x1)
    assert np.array_equal(ds["default"].y, y1)
    # Test the case where x is a single ArrayF and y is an ArrayDict.
    # The keys should come from y.
    ds = Dataset(x1, {"y1": y1, "y2": y2})
    assert np.array_equal(ds["y1"].x, x1)
    assert np.array_equal(ds["y2"].x, x1)
    assert np.array_equal(ds["y1"].y, y1)
    assert np.array_equal(ds["y2"].y, y2)
    # Test the case where x and y are both ArrayDict and keys match.
    # The keys should match between x and y.
    ds = Dataset({"1": x1, "2": x2}, {"1": y1, "2": y2})
    assert np.array_equal(ds["1"].x, x1)
    assert np.array_equal(ds["2"].x, x2)
    assert np.array_equal(ds["1"].y, y1)
    assert np.array_equal(ds["2"].y, y2)
    # Test the case where x and y are both ArrayDict and keys don't match.
    # This should raise a ValueError.
    with pytest.raises(
        ValueError,
        match=re.escape("Keys of 'x', 'y', and 'w' (if w is a dict) must match."),
    ):
        ds = Dataset({"x1": x1, "x2": x2}, {"y1": y1, "y2": y2})
    """
    # TODO:     xx = np.array([6.8, 7.0, 7.2, 7.9])
    with pytest.raises(ValueError, match="Length of 'x' and 'y' must be equal."):
        ds = Dataset(xx, y1)
    ww = np.array([6.8, 7.0, 7.2, 7.9])
    with pytest.raises(ValueError, match="Length of 'x' and 'w' must be equal."):
        ds = Dataset(x1, y1, w=ww)
    """


def test_dataset_copy() -> None:
    """Test deep copy."""
    x1 = np.array([5.5, 7.0, 8.5])
    y1 = np.array([2.1, 1.6, 1.1])
    y2 = np.array([1.8, 1.6, 1.4])
    ds = Dataset(x1, {"1": y1, "2": y2})
    # Test full copy
    ds_copy = ds.copy()
    assert ds_copy.is_ph == ds.is_ph
    assert "1" in ds_copy
    assert "2" in ds_copy
    # Test partial copy
    ds_copy = ds.copy(keys={"1"})
    assert ds_copy.is_ph == ds.is_ph
    assert "1" in ds_copy
    assert "2" not in ds_copy
    # Test KeyError
    with pytest.raises(KeyError):
        ds_copy = ds.copy(keys={"nonexistent"})



"""Test prtecan module."""

from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import Any, ClassVar

import numpy as np
import pandas as pd
import pytest
import seaborn as sns  # type: ignore[import-untyped]
from numpy.testing import assert_almost_equal, assert_array_equal

from clophfit import prtecan
from clophfit.binding.fitting import FitResult
from clophfit.prtecan import (
    Labelblock,
    LabelblocksGroup,
    PlateScheme,
    Tecanfile,
    TecanfilesGroup,
    Titration,
)

# By defining csvl, lb0, and lb1 as class attributes, they are created only once
# per test session. Use fixture to capture UserWarning "OVER"

data_tests = Path(__file__).parent / "Tecan"
pytestmark = pytest.mark.filterwarnings("ignore:OVER")


def test_lookup_listoflines() -> None:
    """It returns indexes for pattern match on col in list of lines."""
    csvl: list[list[str | int | float]] = [
        ["pp", "xy", 1, 2.0],
        ["pp", "xx", 1, 2],
        ["pp", 12, 1, 2],
        ["pp", "yy", 1, 2.0],
        ["a"],
        ["pp", "xy", 1, 2],
    ]
    assert prtecan.lookup_listoflines(csvl, pattern="pp") == [0, 1, 2, 3, 5]
    assert not prtecan.lookup_listoflines(csvl, pattern="xy")  # == []
    assert prtecan.lookup_listoflines(csvl, pattern="xy", col=1) == [0, 5]
    assert prtecan.lookup_listoflines(csvl, pattern="yy", col=1) == [3]
    assert not prtecan.lookup_listoflines(csvl, pattern="yy", col=2)  # == []


def test_strip_lines() -> None:
    """It strips empty fields."""
    lines: list[list[float | int | str]] = [
        ["Excitation Wavelength", "", "", "", 485.0, "nm", "", "", ""]
    ]
    stripped = prtecan.strip_lines(lines)
    assert stripped == [["Excitation Wavelength", 485.0, "nm"]]


def test_extract_metadata() -> None:
    """It extracts metadata correctly."""
    lines: list[list[Any]] = [
        ["Label: Label1", "", "", "", "", "", "", "", "", "", "", "", ""],
        ["Mode", "", "", "", "Fluorescence Top Reading", "", "", "", "", ""],
        ["Shaking (Linear) Amplitude:", "", "", "", 2, "mm", "", "", "", "", ""],
        ["Excitation Wavelength", "", "", "", 400, "nm", "", "unexpected", "", "", ""],
        ["", "Temperature: 26 °C", "", "", "", "", "", "", "", "", ""],
    ]
    expected_metadata = {
        "Label": prtecan.Metadata("Label1"),
        "Mode": prtecan.Metadata("Fluorescence Top Reading"),
        "Shaking (Linear) Amplitude:": prtecan.Metadata(2, ["mm"]),
        "Excitation Wavelength": prtecan.Metadata(400, ["nm", "unexpected"]),
        "Temperature": prtecan.Metadata(26.0, ["°C"]),
    }

    metadata = prtecan.extract_metadata(lines)
    assert metadata == expected_metadata


def test_merge_md() -> None:
    """Merge metadata of both labelblocks and tecanfiles."""
    md1 = {
        "Gain": prtecan.Metadata(93, ["Manual"]),
        "Shaking (Linear) Amplitude:": prtecan.Metadata(2, ["mm"]),
    }
    md2 = {
        "Gain": prtecan.Metadata(93, ["Optimal"]),
        "Shaking (Linear) Amplitude:": prtecan.Metadata(2, ["mm"]),
    }
    mmd = prtecan.merge_md([md1, md2])
    assert mmd["Gain"] == prtecan.Metadata(93)
    assert mmd["Shaking (Linear) Amplitude:"] == prtecan.Metadata(2, ["mm"])


def test_calculate_conc() -> None:
    """Calculates concentration values from Cl additions."""
    additions = [112, 2, 2, 2, 2, 2, 2, 6, 4]
    conc = prtecan.calculate_conc(additions, 1000)
    assert_almost_equal(
        conc, [0.0, 17.544, 34.483, 50.847, 66.667, 81.967, 96.774, 138.462, 164.179], 3
    )


class TestLabelblock:
    """Test labelblock class."""

    @staticmethod
    def _get_two_labelblocks() -> tuple[Labelblock, Labelblock]:
        """Simulate csvl with 2 labelblocks."""
        csvl = prtecan.read_xls(data_tests / "140220/pH6.5_200214.xls")
        idxs = prtecan.lookup_listoflines(csvl)
        # pylint: disable=W0201
        lb0 = Labelblock(csvl[idxs[0] : idxs[1]])
        lb1 = Labelblock(csvl[idxs[1] :])
        return lb0, lb1

    @pytest.fixture(scope="class")
    def labelblocks(self) -> tuple[Labelblock, Labelblock]:
        """Fixture that provides two labelblocks."""
        return self._get_two_labelblocks()

    def test_metadata(self, labelblocks: tuple[Labelblock, Labelblock]) -> None:
        """It parses "Temperature" metadata."""
        lb0, lb1 = labelblocks
        assert lb0.metadata["Temperature"].value == 25.6
        assert lb1.metadata["Temperature"].value == 25.3

    def test_data(self, labelblocks: tuple[Labelblock, Labelblock]) -> None:
        """It parses data values."""
        lb0, lb1 = labelblocks
        assert lb0.data["F06"] == 19551
        assert lb1.data["H12"] == 543

    def test_data_normalized(self, labelblocks: tuple[Labelblock, Labelblock]) -> None:
        """Normalize data using key metadata values."""
        lb0, lb1 = labelblocks
        assert lb0.data_nrm["F06"] == pytest.approx(1051.1290323)
        assert lb1.data_nrm["H12"] == pytest.approx(48.4821429)

    def test_eq(self, labelblocks: tuple[Labelblock, Labelblock]) -> None:
        """A Labelblock is equal to itself and not equal to a different Labelblock."""
        lb0, lb1 = labelblocks
        assert lb0 == lb0  # noqa: PLR0124 # pylint: disable-msg=R0124
        assert lb0 is not lb1
        with pytest.raises(TypeError):
            assert lb0 == 1

    def test_almost_eq(self, labelblocks: tuple[Labelblock, Labelblock]) -> None:
        """Test the __almost_eq__ method of the Labelblock class."""
        lb0, _ = labelblocks
        file_path1 = Path(data_tests) / "290513_7.2.xls"
        csvl1 = prtecan.read_xls(file_path1)  # Gain=98
        idxs1 = prtecan.lookup_listoflines(csvl1)
        lb11 = Labelblock(csvl1[idxs1[1] :])
        file_path2 = Path(data_tests) / "290513_8.8.xls"
        csvl2 = prtecan.read_xls(file_path2)  # Gain=99
        idxs2 = prtecan.lookup_listoflines(csvl2)
        lb12 = Labelblock(csvl2[idxs2[1] :])
        assert lb11 != lb12
        assert lb11.__almost_eq__(lb12)
        assert not lb11.__almost_eq__(lb0)

    def test_overvalue(self, caplog: pytest.LogCaptureFixture) -> None:
        """It detects saturated data ("OVER")."""
        csvl = prtecan.read_xls(data_tests / "140220/pH6.5_200214.xls")
        idxs = prtecan.lookup_listoflines(csvl)
        with caplog.at_level(logging.WARNING):
            lb = Labelblock(csvl[idxs[0] : idxs[1]])
            # Print out the captured logs for debugging
        for log in caplog.records:
            print(log.message)
        expected_messages = [
            " OVER value in Label1: A06 of tecanfile ",
            " OVER value in Label1: H02 of tecanfile ",
        ]
        for expected_message in expected_messages:
            assert any(
                log.message == expected_message for log in caplog.records
            ), f"Expected log message '{expected_message}' not found"
        assert np.nansum(lb.data["A06"]) == np.nansum(np.nan)
        assert np.nansum(lb.data["H02"]) == np.nansum(np.nan)

    def test_raise_missing_column(self) -> None:
        """It raises Exception when a column is missing from the labelblock."""
        csvl = prtecan.read_xls(data_tests / "exceptions/88wells_290212_20.xlsx")
        idxs = prtecan.lookup_listoflines(csvl)
        with pytest.raises(ValueError, match=r"Cannot build Labelblock: not 96 wells?"):
            Labelblock(csvl[idxs[0] : len(csvl)])

    def test_raise_missing_row(self) -> None:
        """It raises Exception when a row is missing from the labelblock."""
        csvl = prtecan.read_xls(data_tests / "exceptions/84wells_290212_20.xlsx")
        idxs = prtecan.lookup_listoflines(csvl)
        with pytest.raises(ValueError, match="Row 7 label mismatch: expected H, got "):
            Labelblock(csvl[idxs[0] : len(csvl)])


class TestCsvlFunctions:
    """Test TecanFile reading and parsing functions."""

    csvl = prtecan.read_xls(data_tests / "140220/pH8.3_200214.xls")

    def test_read_xls(self) -> None:
        """The test reads the xls file using cls method."""
        assert len(self.csvl) == 74

    def test_lookup_listoflines(self) -> None:
        """It finds Label occurrences using module function."""
        assert prtecan.lookup_listoflines(self.csvl) == [14, 44]


class TestTecanfile:
    """Test TecanFile class."""

    tf = prtecan.Tecanfile(data_tests / "140220/pH8.3_200214.xls")

    def test_path(self) -> None:
        """It reads the file path."""
        assert self.tf.path == data_tests / "140220/pH8.3_200214.xls"

    def test_metadata(self) -> None:
        """It parses the Date."""
        assert self.tf.metadata["Date:"].value == "20/02/2014"

    def test_labelblocks(self) -> None:
        """It parses "Temperature" metadata and cell data from 2 labelblocks."""
        assert self.tf.labelblocks[0].metadata["Temperature"].value == 25.3
        assert self.tf.labelblocks[1].metadata["Temperature"].value == 25.7
        assert self.tf.labelblocks[0].data["A01"] == 17260
        assert self.tf.labelblocks[1].data["H12"] == 4196

    def test_eq(self) -> None:
        """A Tecanfile is equal to itself and not equal to a different Tecanfile."""
        tf1 = prtecan.Tecanfile(data_tests / "140220/pH8.3_200214.xls")
        assert self.tf == tf1, "Tecanfile is not equal to itself"
        tf2 = prtecan.Tecanfile(data_tests / "140220/pH9.1_200214.xls")
        assert self.tf != tf2, "Different Tecanfiles are incorrectly reported as equal"

    def test_warn(self, caplog: pytest.LogCaptureFixture) -> None:
        """Warn if labelblocks are repeated in a Tf as it might compromise grouping."""
        with caplog.at_level(logging.WARNING):
            prtecan.Tecanfile(data_tests / "exceptions/290212_7.67_repeated_lb.xls")
        assert any(
            "Repeated labelblocks" in record.message for record in caplog.records
        )

    def test_filenotfound(self) -> None:
        """It raises FileNotFoundError when the file path does not exist."""
        with pytest.raises(FileNotFoundError):
            prtecan.Tecanfile(Path("pinocchio"))

    def test_missing_label(self) -> None:
        """It raises Exception when there is no Label pattern."""
        with pytest.raises(ValueError, match="No Labelblock found."):
            prtecan.Tecanfile(data_tests / "exceptions/0_Labelblocks_290513_5.5.xlsx")


class TestLabelblocksGroup:
    """Test LabelBlocksGroup class."""

    @pytest.fixture(autouse=True, scope="class")
    def tfs(self) -> list[Tecanfile]:
        """Set up list of Tecanfile."""
        return [
            Tecanfile(data_tests / "290513_5.5.xls"),
            Tecanfile(data_tests / "290513_7.2.xls"),
            Tecanfile(data_tests / "290513_8.8.xls"),
        ]

    @pytest.fixture(autouse=True, scope="class")
    def lbgs(self, tfs: list[Tecanfile]) -> tuple[LabelblocksGroup, LabelblocksGroup]:
        """Set up LabelblocksGroup 0 and 1."""
        lbg0 = LabelblocksGroup([tfs[0].labelblocks[0], tfs[1].labelblocks[0]])
        lbg1 = LabelblocksGroup([tfs[1].labelblocks[1], tfs[2].labelblocks[1]])
        return lbg0, lbg1

    def test_metadata(self, lbgs: tuple[LabelblocksGroup, LabelblocksGroup]) -> None:
        """Merge only shared metadata."""
        assert lbgs[0].metadata.get("Temperature") is None
        assert lbgs[1].metadata.get("Temperature") is None
        assert lbgs[1].metadata.get("Gain") is None
        assert lbgs[1].labelblocks[0].metadata["Gain"].value == 98
        assert lbgs[1].labelblocks[1].metadata["Gain"].value == 99
        # Common metadata.
        assert lbgs[0].metadata["Gain"].value == 94
        assert lbgs[0].metadata["Number of Flashes"].value == 10

    def test_data(self, lbgs: tuple[LabelblocksGroup, LabelblocksGroup]) -> None:
        """Merge data."""
        assert lbgs[0].data is not None
        assert lbgs[0].data["A01"] == [18713, 17088]
        assert lbgs[0].data["H12"] == [28596, 25771]
        assert lbgs[1].data == {}

    def test_data_normalized(
        self, lbgs: tuple[LabelblocksGroup, LabelblocksGroup]
    ) -> None:
        """Merge data_normalized."""
        assert_almost_equal(lbgs[1].data_nrm["H12"], [693.980, 714.495], 3)
        assert_almost_equal(lbgs[0].data_nrm["A01"], [995.372, 908.936], 3)

    def test_notequal_labelblocks(self, tfs: list[Tecanfile]) -> None:
        """Raise Exception when concatenating unequal labelblocks."""
        with pytest.raises(ValueError, match="Creation of labelblock group failed."):
            prtecan.LabelblocksGroup([tfs[1].labelblocks[0], tfs[2].labelblocks[1]])


class TestTecanfilesGroup:
    """Group tecanfiles properly."""

    class TestAllEqLbgs:
        """Test TfG with 2 LbG in the same order."""

        @pytest.fixture(autouse=True, scope="class")
        def tfg(self) -> TecanfilesGroup:
            """Set up TecanfilesGroup."""
            filenames = ["290513_5.5.xls", "290513_7.2.xls"]
            tecanfiles = [Tecanfile(data_tests / f) for f in filenames]
            return TecanfilesGroup(tecanfiles)

        def test_metadata(self, tfg: TecanfilesGroup) -> None:
            """Parse general metadata."""
            assert tfg.metadata["Plate"].value == "PE 96 Flat Bottom White   [PE.pdfx]"
            assert tfg.metadata["System"].value == "TECANROBOT"

        def test_labelblocksgroups(self, tfg: TecanfilesGroup) -> None:
            """Generate 2 LbG with .data and .metadata."""
            lbg0 = tfg.labelblocksgroups[0]
            lbg1 = tfg.labelblocksgroups[1]
            # metadata
            assert lbg0.metadata["Number of Flashes"].value == 10.0
            assert lbg1.metadata["Gain"].value == 98.0
            # data normalized ... enough in lbg
            # data
            assert lbg0.data is not None
            assert lbg0.data["A01"] == [18713, 17088]
            assert lbg0.data["H12"] == [28596, 25771]
            assert lbg1.data is not None
            assert lbg1.data["A01"] == [7878, 8761]
            assert lbg1.data["H12"] == [14226, 13602]

    class TestAlmostEqLbgs:
        """Test TfG when 1 LbG equal and a second with almost equal labelblocks."""

        @pytest.fixture
        def tfg_warn(
            self, caplog: pytest.LogCaptureFixture
        ) -> tuple[TecanfilesGroup, list[logging.LogRecord]]:
            """Set up TecanfilesGroup with Warning."""
            filenames = [
                "290513_5.5.xls",  # Label1 and Label2
                "290513_7.2.xls",  # Label1 and Label2
                "290513_8.8.xls",  # Label1 and Label2 with different metadata
            ]
            tecanfiles = [Tecanfile(data_tests / f) for f in filenames]
            with caplog.at_level(logging.WARNING):
                tfg = TecanfilesGroup(tecanfiles)
            return tfg, caplog.records

        def test_log_warning(
            self, tfg_warn: tuple[TecanfilesGroup, list[logging.LogRecord]]
        ) -> None:
            """Warn about labelblocks anomaly."""
            _, records = tfg_warn
            # Check that the specific warning message was logged
            assert any(
                "Different LabelblocksGroup across files" in record.message
                for record in records
            )

        def test_labelblocksgroups(
            self, tfg_warn: tuple[TecanfilesGroup, list[logging.LogRecord]]
        ) -> None:
            """Generate 1 LbG with .data and .metadata."""
            lbg0 = tfg_warn[0].labelblocksgroups[0]
            # metadata
            assert lbg0.metadata["Number of Flashes"].value == 10.0
            assert lbg0.metadata["Gain"].value == 94
            # data
            assert lbg0.data is not None
            assert lbg0.data["A01"] == [18713.0, 17088.0, 17123.0]
            assert lbg0.data["H12"] == [28596.0, 25771.0, 28309.0]

        def test_mergeable_labelblocksgroups(
            self, tfg_warn: tuple[TecanfilesGroup, list[logging.LogRecord]]
        ) -> None:
            """Generate 1 Lbg only with .data_normalized and only common .metadata."""
            lbg1 = tfg_warn[0].labelblocksgroups[1]
            # metadata
            assert lbg1.metadata["Number of Flashes"].value == 10.0
            assert lbg1.metadata.get("Gain") is None
            assert lbg1.data == {}
            # data_normalized
            assert_almost_equal(lbg1.data_nrm["A01"], [401.9387755, 446.9897959, 450.0])
            assert_almost_equal(
                lbg1.data_nrm["H12"], [725.8163265, 693.9795918, 714.4949494]
            )

    class TestOnly1commonLbg:
        """Test TfG with different number of labelblocks, but mergeable."""

        @pytest.fixture
        def tfg_warn(
            self, caplog: pytest.LogCaptureFixture
        ) -> tuple[TecanfilesGroup, list[logging.LogRecord]]:
            """Set up TecanfilesGroup with Warning."""
            filenames = [
                "290212_5.78.xls",  # Label1 and Label2
                "290212_20.xls",  # Label2 only
                "290212_100.xls",  # Label2 only
            ]
            tecanfiles = [Tecanfile(data_tests / f) for f in filenames]
            with caplog.at_level(logging.WARNING):
                tfg = TecanfilesGroup(tecanfiles)
            return tfg, caplog.records

        def test_log_warning(
            self, tfg_warn: tuple[TecanfilesGroup, list[logging.LogRecord]]
        ) -> None:
            """Warn about labelblocks anomaly."""
            _, records = tfg_warn
            # Check that the specific warning message was logged
            assert any(
                "Different LabelblocksGroup across files" in record.message
                for record in records
            )

        def test_labelblocksgroups(
            self, tfg_warn: tuple[TecanfilesGroup, list[logging.LogRecord]]
        ) -> None:
            """Generates 1 LbG with .data and .metadata."""
            tfg, _ = tfg_warn
            lbg = tfg.labelblocksgroups[0]
            # metadata
            assert lbg.metadata["Number of Flashes"].value == 10.0
            assert lbg.metadata["Gain"].value == 93.0
            # data
            assert lbg.data is not None
            assert lbg.data["A01"] == [6289, 6462, 6465]
            assert lbg.data["H12"] == [4477, 4705, 4918]

    class TestFailToMerge:
        """Test TfG without mergeable labelblocks."""

        filenames: ClassVar[list[str]] = ["290513_5.5.xls", "290513_5.5_bad.xls"]
        tecanfiles: ClassVar[list[prtecan.Tecanfile]] = [
            prtecan.Tecanfile(data_tests / f) for f in filenames
        ]

        def test_raise_exception(self) -> None:
            """Raise Exception when there is no way to build labelblocksGroup."""
            msg = r"No common labelblocks in files: ['290513_5.5.xls', '290513_5.5_bad.xls']."
            with pytest.raises(ValueError, match=re.escape(msg)):
                prtecan.TecanfilesGroup(self.tecanfiles)


class TestTitration:
    """Test Titration class."""

    @pytest.fixture
    def tit(self) -> Titration:
        """Set up L1 pH titration."""
        return prtecan.Titration.fromlistfile(data_tests / "L1" / "list.pH", is_ph=True)

    @pytest.fixture(scope="class")
    def tit_ph(self) -> Titration:
        """Set up a pH titration."""
        return prtecan.Titration.fromlistfile(data_tests / "list.pH", is_ph=True)

    @pytest.fixture(scope="class")
    def tit_cl(self) -> Titration:
        """Set up a Cl titration."""
        return prtecan.Titration.fromlistfile(data_tests / "list.cl20", is_ph=False)

    @pytest.fixture(scope="class")
    def tit1(self) -> Titration:
        """Set up a titration with a single Tecan file."""
        tf = prtecan.Tecanfile(data_tests / "140220/pH6.5_200214.xls")
        return prtecan.Titration([tf], conc=np.array([6.5]), is_ph=True)

    def test_conc(self, tit_ph: Titration) -> None:
        """It reads pH values."""
        assert_array_equal(
            tit_ph.conc, [5.78, 6.38, 6.83, 7.24, 7.67, 8.23, 8.82, 9.31]
        )

    def test_labelblocksgroups(self, tit_ph: Titration) -> None:
        """It reads labelblocksgroups data and metadata."""
        lbg0 = tit_ph.labelblocksgroups[0]
        lbg1 = tit_ph.labelblocksgroups[1]
        # metadata
        assert lbg0.metadata["Number of Flashes"].value == 10.0
        # pH9.3 is 93 Optimal not Manual
        assert lbg1.metadata["Gain"] == prtecan.Metadata(93.0)
        # data
        assert lbg0.data is not None
        assert lbg1.data is not None
        assert lbg0.data["A01"][::2] == [30344, 31010, 33731, 37967]
        assert lbg1.data["A01"][1::2] == [9165, 15591, 20788, 22534]
        assert lbg0.data["H12"][1::2] == [20888, 21711, 23397, 25045]
        assert lbg1.data["H12"] == [4477, 5849, 7165, 8080, 8477, 8822, 9338, 9303]

    def test_data_buffersubtracted(self, tit: Titration) -> None:
        """Check data after normalization and bg subtraction."""
        tit.buffer.wells = ["C12", "D01", "D12", "E01", "E12", "F01"]
        tit.params.nrm = False
        assert tit.data[0]
        assert tit.data[1] == {}
        sliced_values = tit.data[0]["B07"][-1::-3][:2]
        assert_almost_equal(sliced_values, [7069, 5716.7], 1)
        # normalization
        tit.params.nrm = True
        sliced_values0 = tit.data[0]["B07"][-1::-3][:2]
        sliced_values1 = tit.data[1]["B07"][-4::-3]
        assert_almost_equal(sliced_values0, [376.01, 304.08], 2)
        assert_almost_equal(sliced_values1, [355.16, 348.57], 2)

    def test_labelblocksgroups_cl(self, tit_cl: Titration) -> None:
        """It reads labelblocksgroups data for Cl too."""
        lbg = tit_cl.labelblocksgroups[0]
        assert lbg.data is not None
        assert lbg.data["A01"] == [6462, 6390, 6465, 6774]
        assert lbg.data["H12"] == [4705, 4850, 4918, 5007]

    def test_export_data(self, tit_ph: Titration, tmp_path: Path) -> None:
        """It exports titrations data to files e.g. "A01.dat"."""
        tit_ph.params.bg = False
        tit_ph.params.dil = False
        tit_ph.params.nrm = False
        tecan_config = prtecan.TecanConfig(
            tmp_path,
            verbose=False,
            comb=False,
            lim=None,
            sel=None,
            title="",
            fit=False,
            png=False,
            pdf=False,
        )
        tit_ph.export_data_fit(tecan_config)
        a01 = pd.read_csv(tmp_path / "dat" / "A01.dat")
        h12 = pd.read_csv(tmp_path / "dat" / "H12.dat")
        assert a01["y1"].tolist()[1::2] == [30072, 32678, 36506, 37725]
        assert a01["y2"].tolist()[1::2] == [9165, 15591, 20788, 22534]
        assert h12["y1"].tolist()[1::2] == [20888, 21711, 23397, 25045]
        assert h12["y2"].tolist() == [4477, 5849, 7165, 8080, 8477, 8822, 9338, 9303]

    def test_raise_listfilenotfound(self) -> None:
        """It raises FileNotFoundError when list.xx file does not exist."""
        with pytest.raises(FileNotFoundError, match="Cannot find: aax"):
            prtecan.Titration.fromlistfile(Path("aax"), True)

    def test_bad_listfile(self) -> None:
        """It raises Exception when list.xx file is ill-shaped."""
        with pytest.raises(ValueError, match=r"Check format .* for listfile: .*"):
            prtecan.Titration.fromlistfile(data_tests / "list.pH2", True)

    def test_data_bg_and_nrm(self, tit1: Titration) -> None:
        """Calculate buffer value from average of buffer wells and subtract."""
        tit1.buffer.wells = ["D01", "D12", "E01", "E12"]
        tit1.params.nrm = False
        tit1.params.dil = False
        tit1.params.bg = False
        tit1.bg = [np.array([11889.25]), np.array([56.75])]
        assert tit1.buffer.dataframes[0]["sem"][0] == pytest.approx(259.9514)
        assert tit1.buffer.dataframes[1]["sem"][0] == pytest.approx(2.561738)
        tit1.params.bg = True
        assert tit1.data[0]["F06"][0] == pytest.approx(7661.75)
        assert tit1.data[1]["H12"][0] == pytest.approx(486.25)
        # Can also assign a buffer value.
        tit1.bg = [np.array([1.0]), np.array([2.9])]
        assert tit1.data[0]["F06"][0] == 19550
        assert tit1.data[1]["H12"][0] == 540.1
        # nrm
        assert tit1.buffer.dataframes_nrm[0]["fit"][0] == pytest.approx(639.20699)
        assert tit1.buffer.dataframes_nrm[0]["mean"][0] == pytest.approx(639.20699)
        assert tit1.buffer.dataframes_nrm[1]["fit"][0] == pytest.approx(5.06696)
        assert tit1.buffer.dataframes_nrm[0]["sem"][0] == pytest.approx(13.97588)
        assert tit1.buffer.dataframes_nrm[1]["sem"][0] == pytest.approx(0.2287266)
        # also bg duplicates data in buffers_nrm
        tit1.params.nrm = True
        tit1.buffer.wells = ["D01", "D12", "E01", "E12"]
        assert tit1.bg[0][0] == pytest.approx(639.20699)
        assert tit1.bg[1] == pytest.approx(5.06696)
        # nrm data
        assert tit1.data[0]["F06"] == pytest.approx(411.922)
        assert tit1.data[1]["H12"] == pytest.approx(43.4152)
        # Can also assign a buffer_norm value.
        tit1.bg = [np.array([1.0]), np.array([0.4821])]
        assert tit1.data[0]["F06"] == pytest.approx(1050.13)
        assert tit1.data[1]["H12"] == pytest.approx(48.0)


class TestPlateScheme:
    """Test PlateScheme."""

    @pytest.fixture(autouse=True, scope="class")
    def ps(self) -> PlateScheme:
        """Create a void PlateScheme."""
        return PlateScheme()

    def test_buffer(self, ps: PlateScheme) -> None:
        """Set buffer and test raise error."""
        ps.buffer = ["A1", "A2"]
        assert ps.buffer == ["A1", "A2"]
        with pytest.raises(TypeError):
            ps.buffer = [1, 2]  # type: ignore[list-item]

    def test_ctrl(self, ps: PlateScheme) -> None:
        """Set ctrl and test raise error."""
        ps.ctrl = ["B1", "B2"]
        assert ps.ctrl == ["B1", "B2"]
        with pytest.raises(TypeError):
            ps.ctrl = [1, 2]  # type: ignore[list-item]

    def test_names(self, ps: PlateScheme) -> None:
        """Set names and test raise error."""
        ps.names = {"name1": {"A1", "A2"}, "name2": {"B1", "B2"}}
        assert ps.names == {"name1": {"A1", "A2"}, "name2": {"B1", "B2"}}
        with pytest.raises(TypeError):
            ps.names = {"name1": [1, 2], "name2": [3, 4]}  # type: ignore[dict-item]

    def test_invalid_file(self) -> None:
        """Test providing an incorrect file."""
        with pytest.raises(FileNotFoundError):
            PlateScheme(file=Path("incorrect_file.csv"))


# some:  @pytest.mark.skipif(sys.platform == "win32", reason="broken on windows")
class TestTitrationAnalysis:
    """Test TitrationAnalysis class."""

    @pytest.fixture(autouse=True, scope="class")
    def titan(self) -> Titration:
        """Set up TitrationAnalysis."""
        titan = Titration.fromlistfile(data_tests / "140220/list.pH", is_ph=True)
        titan.load_additions(data_tests / "140220/additions.pH")
        titan.load_scheme(data_tests / "140220/scheme.txt")
        return titan

    @pytest.fixture(autouse=True, scope="class")
    def titan_no_scheme(self) -> Titration:
        """Set up TitrationAnalysis without scheme."""
        return Titration.fromlistfile(data_tests / "140220/list.pH", is_ph=True)

    @pytest.fixture(autouse=True, scope="class")
    def titan_1lbg(self) -> Titration:
        """Set up TitrationAnalysis with only 1 lbg before normalization."""
        titan = Titration.fromlistfile(data_tests / "L1/list.pH", is_ph=True)
        titan.load_additions(data_tests / "L1/additions.pH")
        titan.load_scheme(data_tests / "L1/scheme.txt")
        return titan

    def test_scheme(self, titan: Titration) -> None:
        """It finds well position for buffer samples."""
        assert titan.scheme.buffer == ["D01", "E01", "D12", "E12"]
        assert titan.buffer.wells == ["D01", "E01", "D12", "E12"]

    def test_raise_listfilenotfound(self, titan: Titration) -> None:
        """It raises OSError when scheme file does not exist."""
        with pytest.raises(
            FileNotFoundError, match=r"No such file or directory: 'aax'"
        ):
            titan.load_scheme(Path("aax"))

    def test_raise_listfile_exception(self, titan: Titration) -> None:
        """It raises AssertionError when scheme.txt file is ill-shaped."""
        bad_schemefile = data_tests / "140220/scheme0.txt"
        msg = f"Check format [well sample] for schemefile: {bad_schemefile}"
        with pytest.raises(ValueError, match=re.escape(msg)):
            titan.load_scheme(bad_schemefile)

    def test_subtract_bg(self, titan: Titration) -> None:
        """It subtracts buffer average values."""
        lbg0 = titan.labelblocksgroups[0]
        lbg1 = titan.labelblocksgroups[1]
        assert_almost_equal(
            lbg0.data_nrm["E01"][::2], [601.72, 641.505, 674.355, 706.774], 3
        )
        assert lbg0.data is not None
        assert lbg0.data["E01"][::2] == [11192.0, 11932.0, 12543.0, 13146.0]
        assert type(lbg1) is LabelblocksGroup
        titan.params.bg = True
        titan.params.nrm = False
        titan.params.dil = False
        assert_array_equal(titan.data[0]["A12"][::3], [8084.5, 16621.75, 13775.0])
        assert lbg1.data is not None
        assert_array_equal(titan.data[1]["A12"][::3], [9758.25, 1334.0, 283.5])

    def test_dilution_correction(self, titan: Titration) -> None:
        """It applies dilution correction read from file listing additions."""
        assert titan.additions is not None
        assert_array_equal(titan.additions, [100, 2, 2, 2, 2, 2, 2])
        titan.params.nrm = False
        titan.params.dil = True
        assert titan.data is not None
        assert titan.data[1] is not None
        assert_almost_equal(
            titan.data[1]["A12"],
            [9758.25, 7524.795, 3079.18, 1414.04, 641.79, 402.325, 317.52],
        )

    def test_data_nrm(self, titan: Titration) -> None:
        """It normalizes data."""
        titan.params.nrm = True
        titan.params.bg = True
        titan.params.dil = True

        assert_almost_equal(
            titan.data[0]["A12"][::2],
            [434.65, 878.73, 975.58, 829.46],
            2,
        )
        assert_almost_equal(
            titan.data[1]["A12"][::2],
            [871.272, 274.927, 57.303, 28.35],
            3,
        )

    def test_keys(self, titan: Titration) -> None:
        """It gets well positions for ctrl and unknown samples."""
        assert set(titan.scheme.names) == {"NTT", "G03", "V224Q", "S202N"}
        x = {"B12", "H12", "F01", "C12", "F12", "C01", "H01", "G12", "B01", "G01"}
        assert set(titan.scheme.ctrl) - {"A01", "A12"} == x

    def test_fit(self, titan: Titration) -> None:
        """It fits each label separately."""
        fres = titan.results
        # Check that the first fit result dictionary has 92 elements
        assert len(fres[0]) == 92
        # Check that the first fit result for 'H02' is None
        assert fres[0]["H02"] == FitResult(None, None, None)
        # Check that the second fit result for 'H02' is not None
        assert fres[1]["H02"].is_valid()
        # Check 'K' and std error for 'H02' in the second fit result
        assert fres[1]["H02"].result is not None
        k_h02 = fres[1]["H02"].result.params["K"]
        assert k_h02.value == pytest.approx(7.8904, abs=1e-4)
        assert k_h02.stderr == pytest.approx(0.0170, abs=1e-4)
        # Check 'K' and std error for 'H02' in the third fit result
        assert fres[2]["H02"].result is not None
        k_h02 = fres[2]["H02"].result.params["K"]
        assert k_h02.value == pytest.approx(7.8904, abs=1e-4)
        assert k_h02.stderr == pytest.approx(0.0169, abs=1e-4)
        # Check 'K' and std error for 'E02' in the second fit result
        assert fres[1]["E02"].result is not None
        k_e02 = fres[1]["E02"].result.params["K"]
        assert k_e02.value == pytest.approx(7.9771, abs=1e-4)
        assert k_e02.stderr == pytest.approx(0.0243, abs=1e-4)
        # Check 'K' and std error for 'E02' in the third fit result
        assert fres[2]["E02"].result is not None
        k_e02 = fres[2]["E02"].result.params["K"]
        assert k_e02.value == pytest.approx(7.9771, abs=1e-4)
        assert k_e02.stderr == pytest.approx(0.0169, abs=1e-4)
        # Fit up to the second-last data point

    def test_plot_buffer_with_title(self, titan: Titration) -> None:
        """It plots buffers for 2 lbg with title."""
        g = titan.buffer.plot(title="Test Title")
        assert isinstance(g, sns.FacetGrid)
        assert len(g.axes_dict) == 2
        assert g.fig._suptitle.get_text() == "Test Title"  # noqa: SLF001

    def test_plot_buffer_normalized(self, titan: Titration) -> None:
        """It plots buffers_norm for 2 lbg."""
        g = titan.buffer.plot(nrm=True)
        assert isinstance(g, sns.FacetGrid)
        assert len(g.axes_dict) == 2

    def test_plot_buffer_empty_buffers(self, titan_no_scheme: Titration) -> None:
        """It handles empty buffers (before assignment of buffer_wells)."""
        g = titan_no_scheme.buffer.plot()
        assert isinstance(g, sns.FacetGrid)

    def test_plot_buffer_1lbg(self, titan_1lbg: Titration) -> None:
        """It plots buffers_norm in the case of 1 mergeable lbg."""
        g = titan_1lbg.buffer.plot()
        assert isinstance(g, sns.FacetGrid)
        assert len(g.axes_dict) == 1

    def test_plot_buffer_1lbg_normalized(self, titan_1lbg: Titration) -> None:
        """It plots buffers_norm in the case of 1 mergeable lbg."""
        g = titan_1lbg.buffer.plot(nrm=True)
        assert isinstance(g, sns.FacetGrid)
        assert len(g.axes_dict) == 2



"""Parse Tecan files and fit files that are grouped into titrations.

A 'titration' in this context is defined within a list.pH or list.cl file. These
files describe the properties and parameters of the titration experiment.

The command-line interface 'prtecan' is used to construct 96 distinct titrations
and export the results to .txt files. This allows for further analysis or
visualization of the titration data.

In cases where two label blocks exist in the Tecan files, the module performs a
'global fitting' operation. Global fitting is a method used in data analysis to
best fit a model to multiple data sets simultaneously, under the assumption that
some underlying parameters are shared between the data sets.

Please consult the documentation for each function in this module for more
detailed information on their operation and use.
"""

from clophfit.prtecan.prtecan import (
    Labelblock,
    LabelblocksGroup,
    Metadata,
    PlateScheme,
    TecanConfig,
    Tecanfile,
    TecanfilesGroup,
    Titration,
    TitrationConfig,
    TitrationPlotter,
    calculate_conc,
    dilution_correction,
    extract_metadata,
    lookup_listoflines,
    merge_md,
    read_xls,
    strip_lines,
)

__all__ = [
    "FitdataParams",
    "Labelblock",
    "LabelblocksGroup",
    "Metadata",
    "PlateScheme",
    "TecanConfig",
    "Tecanfile",
    "TecanfilesGroup",
    "Titration",
    "TitrationConfig",
    "TitrationPlotter",
    "calculate_conc",
    "dilution_correction",
    "extract_metadata",
    "lookup_listoflines",
    "merge_md",
    "read_xls",
    "strip_lines",
]



"""Test prenspire module."""

import warnings
from pathlib import Path

import pytest

from clophfit.prenspire import EnspireFile, Note

data_files_dir = Path(__file__).parent / "EnSpire"
esff = data_files_dir.joinpath


@pytest.fixture(scope="module", name="ef1")
def ef1_setup() -> EnspireFile:
    """Read in file."""
    return EnspireFile(esff("h148g-spettroC.csv"))


@pytest.fixture(scope="module", name="ef2")
def ef2_setupt() -> EnspireFile:
    """Read in file without 'Samples' column."""
    return EnspireFile(esff("e2-T-without_sample_column.csv"))


class TestEnspireFile:
    """Test EnspireFile class."""

    def test_exceptions(self) -> None:
        """Test some raised exceptions."""
        esf = (data_files_dir / "exceptions").joinpath
        # Test get_data_ini Exceptions
        with pytest.raises(Exception, match="No line starting with ."):
            EnspireFile(esf("h148g-spettroC-idx0.csv"))
        with pytest.raises(Exception, match="Multiple lines starting with ."):
            EnspireFile(esf("h148g-spettroC-idx2.csv"))
        # Test platemap
        with pytest.raises(Exception, match="stop: Platemap format unexpected"):
            EnspireFile(esf("M1-A6-G12_11columns.csv"))
        # Test the presence of some empty lines
        with pytest.raises(Exception, match="Expecting two empty lines before _ini"):
            EnspireFile(esf("M1-A6-G12_missing_emptyline_ini.csv"))
        with pytest.raises(Exception, match="Expecting an empty line after _fin"):
            EnspireFile(esf("M1-A6-G12_missing_emptyline_fin.csv"))
        # Test multiple emission wavelengths in excitation spectra
        with pytest.raises(
            Exception, match="Excitation spectra with unexpected emission in MeasA"
        ):
            EnspireFile(esf("e2dan-exwavelength.csv"))
        with pytest.raises(
            Exception, match="Excitation spectra with unexpected emission in MeasA"
        ):
            EnspireFile(esf("e2dan-exwavelength2.csv"))
        with pytest.raises(
            Exception, match="Emission spectra with unexpected excitation in MeasC"
        ):
            EnspireFile(esf("e2dan-emwavelength.csv"))
        with pytest.raises(
            Exception, match='Unknown "Monochromator": Strange in MeasB'
        ):
            EnspireFile(esf("e2dan-exwavelengthstrange.csv"))

    def test_get_data_ini(self, ef1: EnspireFile, ef2: EnspireFile) -> None:
        """Test get_data_ini."""
        assert ef1._ini == 12  # noqa: SLF001 # pylint: disable-msg=W0212
        assert ef2._ini == 9  # noqa: SLF001 # pylint: disable-msg=W0212

    def test_fin(self, ef1: EnspireFile, ef2: EnspireFile) -> None:
        """Test _fin (line_index())."""
        assert ef1._fin == 14897  # noqa: SLF001 # pylint: disable-msg=W0212
        assert ef2._fin == 856  # noqa: SLF001 # pylint: disable-msg=W0212

    """
    # def test_metadata_post(self, ef1: EnspireFile, ef2: EnspireFile) -> None:
    #     \"Identify correctly the beginning of metadata after data block.\"
    #     assert ef1._metadata_post[0] == ["Basic assay information "]
    #     assert ef2._metadata_post[0] == ["Basic assay information "]

    # def test_localegen_in_metadata_post(
    #     self, ef1: EnspireFile, ef2: EnspireFile
    # ) -> None:
    #     \"Test locales.\"
    #     assert ef1._metadata_post[31][4] == "300 µl"
    #     assert ef2._metadata_post[31][4] == "300 µl"

    # def test_data_list(self, ef1: EnspireFile, ef2: EnspireFile) -> None:
    #     \"Test data_list.\"
    #     assert ef1._data_list[0][2] == "MeasA:Result"
    #     assert ef2._data_list[0][2] == "MeasB:WavelengthEms"
    #     assert ef1._data_list[1][2] == "3151"
    #     assert ef2._data_list[1][3] == "3739"
    #     # Important to check for last line
    #     assert ef1._data_list[-1][2] == "97612"
    #     assert ef2._data_list[-1][3] == "512"
    """

    def test_get_list_from_platemap(self, ef1: EnspireFile, ef2: EnspireFile) -> None:
        """Test list from platemap."""
        assert ef1._wells_platemap[2] == "A03"  # noqa: SLF001 # pylint: disable-msg=W0212
        assert ef2._wells_platemap[1] == "F02"  # noqa: SLF001 # pylint: disable-msg=W0212

    def test_metadata(self, ef1: EnspireFile) -> None:
        """Test metadata dictionary."""
        assert ef1.metadata["Measurement date"] == "2011-10-03 17:12:33"
        assert ef1.metadata["Chamber temperature at start"] == "20"
        assert ef1.metadata["Chamber temperature at end"] == "20"
        assert ef1.metadata["Ambient temperature at start"] == "6"
        assert ef1.metadata["Ambient temperature at end"] == "9.3"
        assert ef1.metadata["Protocol name"] == "Eccitazione C"
        assert (
            ef1.metadata["Exported data"] == "Well,Sample,MeasA:Result,MeasA:Wavelength"
        )

    def test_measurement_metadata(self, ef1: EnspireFile, ef2: EnspireFile) -> None:
        """Test data object."""
        assert ef1.measurements["A"]["metadata"]["Monochromator"] == "Excitation"
        assert ef1.measurements["A"]["metadata"]["Wavelength"] == "520"
        assert ef1.measurements["A"]["metadata"]["temp"] == "20"
        assert ef2.measurements["C"]["metadata"]["Monochromator"] == "Emission"
        assert ef2.measurements["C"]["metadata"]["Wavelength"] == "480"
        assert ef2.measurements["F"]["metadata"]["temp"] == "35"

    def test_measurements(self, ef1: EnspireFile, ef2: EnspireFile) -> None:
        """Test data object."""
        assert ef1.measurements["A"]["lambda"][0] == 272
        assert ef1.measurements["A"]["lambda"][228] == 500
        assert ef1.measurements["A"]["A01"][0] == 3151
        assert ef1.measurements["A"]["A01"][228] == 573
        # Important to check for last line of csvl
        assert ef1.measurements["A"]["G05"][228] == 97612
        assert ef1.measurements["A"]["A01"][10] == 2496
        assert ef1.measurements["A"]["A02"][10] == 1765
        assert ef1.measurements["A"]["lambda"][10] == 282
        assert ef1.measurements["A"]["metadata"]["Wavelength"] == "520"
        assert ef2.measurements["A"]["lambda"][2] == 402
        assert ef2.measurements["B"]["F07"][1] == 1898
        assert ef2.measurements["A"]["metadata"]["Wavelength"] == "530"

    def test_check_lists_warning(self) -> None:
        """It warns when (csv != platemap).

        *-incomplete.csv: 5 wells (G1-G5) are missing. (They are present in
        Platemap because they were present during acquisition and then not
        exported).
        """
        esf = (data_files_dir / "warnings").joinpath

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            EnspireFile(esf("h148g-spettroC-incomplete.csv"))
            assert len(w) == 2
            assert issubclass(w[-1].category, Warning)
            assert "platemap" in str(w[-1].message)

    # placemark def test_get_maxx(self):
    # placemark    "really need to be completed"
    # placemark    self.assertEqual(self.s.get_maxx(self.s.ex, self.s.y), 272)


@pytest.fixture(scope="module", name="n1")
def n1_setup() -> Note:
    """Read note file."""
    return Note(esff("h148g-spettroC-nota.csv"))


@pytest.fixture(scope="module", name="n1err")
def n1err_setup() -> Note:
    """Read note file with missing wells."""
    return Note(esff("h148g-spettroC-nota-Err.csv"))


class TestNote:
    """Experimental notes."""

    def test_get_list_from_note(self, n1: Note) -> None:
        """Test get_well_list_from_note method."""
        assert n1.wells[2] == "A03"

    def test__note_list(self, n1: Note) -> None:
        """Test well_list from note."""
        assert n1._note.loc[2, "Well"] == "A03"  # noqa: SLF001 # pylint: disable-msg=W0212
        assert n1._note.iloc[64, 1] == 8.2  # noqa: SLF001 # pylint: disable-msg=W0212

    def test_wells(self, n1: Note, ef1: EnspireFile, n1err: Note) -> None:
        """Check wells from Note vs EnspireFile."""
        assert n1.wells == ef1.wells
        assert (n1err.wells == ef1.wells) is False

    def test_build_titrations(self, n1: Note, ef1: EnspireFile) -> None:
        """Test the method extract_titrations()."""
        n1.build_titrations(ef1)
        assert len(n1.titrations["H148G"][20.0]) == 16
        tit0 = n1.titrations["H148G"][20]["Cl_0.0"]["A"]
        assert tit0.columns.to_list() == [5.2, 6.3, 7.4, 8.1, 8.2]
        assert tit0[5.2][272] == 3151
        tit5 = n1.titrations["H148G"][20]["pH_8.2"]["A"]
        assert tit5[667.0][500] == 8734



"""ClopHfit: Parse plate-reader and fit ClopHensor titrations."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("clophfit")
except PackageNotFoundError:
    __version__ = "unknown"

__enspire_out_dir__ = f"Meas-{__version__}"
__tecan_out_dir__ = f"out-{__version__}"



"""Test ``clop`` cli."""

import csv
import filecmp
import logging
import re
import tempfile
from pathlib import Path
from typing import IO, cast

import pytest
from click.testing import CliRunner
from matplotlib.testing.compare import compare_images
from matplotlib.testing.exceptions import ImageComparisonFailure

from clophfit.__main__ import clop, fit_titration, note2csv, ppr

# tests path
tpath = Path(__file__).parent


@pytest.fixture(name="runner")
def runner_setup() -> CliRunner:
    """Fixture for invoking command-line interfaces."""
    return CliRunner()


def create_temp_tsv_file(content: str) -> IO[str]:
    """Create a temporary TSV file and populate it with content."""
    with tempfile.NamedTemporaryFile(
        delete=False, mode="w+", suffix=".tsv"
    ) as temp_file:
        temp_file.write(content)
        temp_file.seek(0)
        return cast(IO[str], temp_file)


def test_default_case(runner: CliRunner) -> None:
    """Test default case for note2csv function."""
    temp_file = create_temp_tsv_file("Well\tpH\tCl\tName\nD01\t9.15\t0\tNTT-G10\n")
    result = runner.invoke(note2csv, [temp_file.name])
    assert result.exit_code == 0
    # Read the output CSV and check its contents
    output_path = Path(temp_file.name).with_suffix(".csv")
    with output_path.open("r", encoding="utf-8") as f:
        reader = csv.reader(f)
        header = next(reader)
        assert header == ["Well", "pH", "Cl", "Name", "Temp", "Labels"]
        row = next(reader)
        assert row == ["D01", "9.15", "0", "NTT-G10", "37.0", "A B"]


def test_custom_output(runner: CliRunner) -> None:
    """Test custom case for note2csv function."""
    temp_file = create_temp_tsv_file("Well\tpH\tCl\tName\nD01\t9.15\t0\tNTT-G10\n")

    with tempfile.NamedTemporaryFile(
        delete=False, mode="w+", suffix=".csv"
    ) as output_file:
        output_file.write("Well,pH,Cl,Name,Temp,Labels" + "\n")
        output_file.seek(0)  # Important to rewind the file to the beginning
        result = runner.invoke(
            note2csv,
            [temp_file.name, "-o", output_file.name, "-t", "22.3", "-l", "A F G"],
        )

    assert result.exit_code == 0

    # Read the output CSV and check its contents
    output_path = Path(output_file.name)
    with output_path.open("r", encoding="utf-8") as f:
        reader = csv.reader(f)
        header = next(reader)
        assert header == ["Well", "pH", "Cl", "Name", "Temp", "Labels"]
        row = next(reader)
        assert row == ["D01", "9.15", "0", "NTT-G10", "22.3", "A F G"]


def test_eq1() -> None:
    """It runs XXX pr.tecan and generates correct results and graphs."""
    runner = CliRunner()
    result = runner.invoke(clop, ["eq1", "2", "2", "2"])
    assert result.exit_code == 0
    assert "4." in result.output


def test_prenspire(tmp_path: Path) -> None:
    """Test prenspire command with actual data and validate output."""
    expected = tpath / "EnSpire" / "cli" / "output"
    input_csv = tpath / "EnSpire" / "cli" / "NTT_37C_pKa.csv"
    out = tmp_path / "E"
    out.mkdir()
    runner = CliRunner()
    result = runner.invoke(ppr, ["--out", str(out), "enspire", str(input_csv)])
    assert result.exit_code == 0
    # validate output files
    assert (out / "NTT_37C_pKa_A.csv").exists()
    assert (out / "NTT_37C_pKa_B.csv").exists()
    assert (out / "NTT_37C_pKa_A.png").exists()
    assert (out / "NTT_37C_pKa_B.png").exists()
    # validate output file contents
    assert filecmp.cmp(out / "NTT_37C_pKa_A.csv", expected / "NTT_37C_pKa_A.csv")
    assert filecmp.cmp(out / "NTT_37C_pKa_B.csv", expected / "NTT_37C_pKa_B.csv")
    # validate graph
    for f in ["NTT_37C_pKa_A.png", "NTT_37C_pKa_B.png"]:
        msg = compare_images(str(out / f), str(expected / f), 0.007)
        if msg:  # pragma: no cover
            raise ImageComparisonFailure(msg)


def test_prtecan(tmp_path: Path, caplog: pytest.LogCaptureFixture) -> None:
    """Test prtecan command with actual data."""
    list_f = str(tpath / "Tecan" / "140220" / "list.pH")
    scheme_f = str(tpath / "Tecan" / "140220" / "scheme.txt")
    out = str(tmp_path)
    runner = CliRunner()
    # Ensure we're capturing warnings at the appropriate level
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            ppr,
            ["--out", out, "tecan", list_f, "--fit", "--sch", scheme_f, "--bg"],
        )
    assert any("OVER" in record.message for record in caplog.records), "OVER not found"
    assert result.exit_code == 0


def test_prtecan_cl(tmp_path: Path) -> None:
    """Test prtecan command with actual data."""
    list_f = str(tpath / "Tecan" / "140220" / "list.cl")
    scheme_f = str(tpath / "Tecan" / "140220" / "scheme.txt")
    adds_f = str(tpath / "Tecan" / "140220" / "additions.cl")
    out = tmp_path / "out4"
    out.mkdir()
    runner = CliRunner()
    base_args = ["--out", str(out), "tecan", list_f, "--no-fit", "--sch", scheme_f]
    base_args.extend(["--add", adds_f, "--bg", "--cl", "1000.0"])
    result = runner.invoke(ppr, base_args)
    assert result.exit_code == 0


@pytest.mark.parametrize(
    ("csv_file", "output", "opts"),
    [
        ("A04 Cl_A.csv", " K :  13.64  14.28  14.86  15.44  16.04  16.71  17.51", None),
        ("A04 Cl_A.csv", " K :  13.52  14.11  14.66  15.20", ["-b", "480", "530"]),
        ("A04 Cl_B.csv", " K :  12.73  13.46  14.14  14.82  15.54  16.35  17.32", None),
        ("A04 Cl_B.csv", " K :  12.93  13.66  14.34  15.02", ["-b", "480", "530"]),
    ],
)
def test_fit_titration(
    csv_file: str, output: str, opts: list[str], tmp_path: Path
) -> None:
    """Test the old ``fit_titration.py`` script with actual data and validate output."""
    note_fp = tpath / "spec" / "NTT-A04-Cl_note"
    csv_fp = tpath / "spec" / "Meas" / csv_file
    out = tmp_path / "tit_fit"
    out.mkdir()
    runner = CliRunner()
    base_args = ["--no-is-ph", "-o", str(out), "spec", str(csv_fp), str(note_fp)]
    if opts:
        base_args.extend(opts)
        sbands = f"({opts[1]}, {opts[2]})"
    else:
        sbands = "None"
    result = runner.invoke(fit_titration, base_args)
    assert result.exit_code == 0
    expected_output = re.sub(" ", r"\\s+", output.strip())
    # Assert that the pattern appears in the output
    assert re.search(expected_output, result.output) is not None
    # Asserting that PDF is created
    expected_pdf_filename = out / f"{csv_fp.stem}_{sbands}_{note_fp.stem}.pdf"
    assert expected_pdf_filename.exists()


@pytest.mark.parametrize(
    ("dat_f", "ph_opt", "output", "opts"),
    [
        ("pH_D05.dat", "--is-ph", "K: 7.5313", ["-b", "78"]),
        ("pH_D05.dat", "--is-ph", "K: 7.65166", ["--no-weight"]),
        ("Cl_B05-20130628-cor.dat", "--no-is-ph", "K: 6.4537", None),
    ],
)
def test_fit_titration_glob(dat_f: str, ph_opt: str, output: str, opts: str) -> None:
    """Fit K correctly, generate png files as the old ``fit_titration_global.py``."""
    dat_fp = tpath / "glob" / dat_f
    base_args = ["-v", ph_opt, "glob", str(dat_fp)]
    if opts:
        base_args.extend(opts)
    runner = CliRunner()
    result = runner.invoke(fit_titration, base_args)
    assert result.exit_code == 0
    expected_output = re.sub(" ", r"\\s+", output.strip())
    assert re.search(expected_output, result.output) is not None
    # assert that the png files are generated
    pngs = [dat_fp.with_suffix(".png")]
    if opts and "-b" in opts:
        pngs.append(dat_fp.with_stem(dat_fp.stem + "-emcee").with_suffix(".png"))
        if "--is-ph" in ph_opt:
            pngs.append(
                dat_fp.with_stem(dat_fp.stem + "-emc-ratios").with_suffix(".png")
            )
    for file in pngs:
        assert file.exists(), f"{file} does not exist!"
    # delete the png files
    for file in pngs:
        file.unlink()



"""Type definitions used throughout the clophfit package."""

import numpy as np
from numpy.typing import NDArray

ArrayF = NDArray[np.float64]
ArrayDict = dict[str, ArrayF]
Kwargs = dict[str, str | int | float | bool | None]



"""Fit Cl binding and pH titration."""

from __future__ import annotations

import copy
import typing
import warnings
from dataclasses import dataclass, field
from sys import float_info

import lmfit  # type: ignore[import-untyped]
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from lmfit import Parameters
from lmfit.minimizer import Minimizer, MinimizerResult  # type: ignore[import-untyped]
from matplotlib import axes, figure
from scipy import odr  # type: ignore[import-untyped]
from uncertainties import ufloat  # type: ignore[import-untyped]

from clophfit.binding.plotting import (
    COLOR_MAP,
    PlotParameters,
    _apply_common_plot_style,
    _create_spectra_canvas,
    plot_autovalues,
    plot_autovectors,
    plot_emcee_k_on_ax,
    plot_pca,
    plot_spectra,
    plot_spectra_distributed,
)
from clophfit.types import ArrayDict, ArrayF

N_BOOT = 20  # To compute fill_between uncertainty.
EMCEE_STEPS = 1800


@dataclass
class DataArrays:
    """A collection of matching x, y, and optional w data arrays."""

    x: ArrayF
    y: ArrayF
    w: ArrayF | None = None

    def __post_init__(self) -> None:
        """Ensure the x and y arrays are of equal length after initialization."""
        if len(self.x) != len(self.y):
            msg = "Length of 'x' and 'y' must be equal."
            raise ValueError(msg)
        if self.w is not None and len(self.x) != len(self.w):
            msg = "Length of 'x' and 'w' must be equal."
            raise ValueError(msg)


class Dataset(dict[str, DataArrays]):
    """Contain pairs of matching x and y data arrays, indexed by a string key.

    Initialization will remove any NaN values from the x and y arrays.

    Parameters
    ----------
    x : ArrayF | ArrayDict
        The x values of the dataset(s), either as a single ArrayF or as an ArrayDict
        if multiple datasets are provided.
    y : ArrayF | ArrayDict
        The y values of the dataset(s), either as a single ArrayF or as an ArrayDict
        if multiple datasets are provided.
    is_ph : bool
        Indicate if x values represent pH (default is False).
    w : ArrayF | ArrayDict | None
        The w values (weights) of the dataset(s), either as a single ArrayF or
        as an ArrayDict if multiple datasets are provided.

    Raises
    ------
    ValueError
        If x and y are both ArrayDict and their keys don't match.
    """

    is_ph: bool = False

    def __init__(
        self,
        x: ArrayF | ArrayDict,
        y: ArrayF | ArrayDict,
        is_ph: bool = False,
        w: ArrayF | ArrayDict | None = None,
    ) -> None:
        def filter_nan(arr: ArrayF, reference: ArrayF) -> ArrayF:
            return typing.cast(ArrayF, arr[~np.isnan(reference)])

        def create_data(x: ArrayF, y: ArrayF, w: ArrayF | None) -> DataArrays:
            x, y = filter_nan(x, y), filter_nan(y, y)
            w = filter_nan(w, y) if w is not None else None
            return DataArrays(x, y, w)

        # x:array, y:array
        if (
            isinstance(x, np.ndarray)
            and isinstance(y, np.ndarray)
            and not isinstance(w, dict)
        ):
            data = {"default": create_data(x, y, w)}
        # x:array, y:dict_of_arrays
        elif isinstance(x, np.ndarray) and isinstance(y, dict):
            data = {
                k: create_data(x, v, w[k] if isinstance(w, dict) else w)
                for k, v in y.items()
            }
        # x:dict_of_arrays, y:dict_of_arrays
        elif isinstance(x, dict) and isinstance(y, dict):
            if x.keys() != y.keys() or (isinstance(w, dict) and x.keys() != w.keys()):
                msg = "Keys of 'x', 'y', and 'w' (if w is a dict) must match."
                raise ValueError(msg)
            data = {
                k: create_data(x[k], y[k], w[k] if isinstance(w, dict) else w)
                for k in x
            }
        super().__init__(data)
        self.is_ph = is_ph

    def add_weights(self, w: ArrayF | ArrayDict) -> None:
        """Add weights to the dataset.

        Parameters
        ----------
        w : ArrayF | ArrayDict
            The weights to be added to the dataset.

        Raises
        ------
        ValueError
            If a key in the weights dictionary does not match any key in the
            current Dataset object.
        """
        if isinstance(w, np.ndarray):
            for da in self.values():
                da.w = w
        elif isinstance(w, dict):
            for k, weights in w.items():
                if k in self:
                    self[k].w = weights
                else:
                    msg = f"No matching found for key '{k}' in the current Dataset."
                    raise ValueError(msg)

    def copy(self, keys: list[str] | set[str] | None = None) -> Dataset:
        """Return a copy of the Dataset.

        If keys are provided, only data associated with those keys are copied.

        Parameters
        ----------
        keys : list[str] | set[str] | None, optional
            List of keys to include in the copied dataset. If None (default),
            copies all data.

        Returns
        -------
        Dataset
            A copy of the dataset.

        Raises
        ------
        KeyError
            If a provided key does not exist in the Dataset.
        """
        if keys is None:
            copied = copy.deepcopy(self)
        else:
            # If keys are specified, only copy those keys
            copied = Dataset({}, {}, is_ph=self.is_ph)
            for key in keys:
                if key in self:
                    copied[key] = copy.deepcopy(self[key])
                else:
                    msg = f"No such key: '{key}' in the Dataset."
                    raise KeyError(msg)
        return copied

    def clean_data(self, n_params: int) -> None:
        """Remove too small datasets."""
        for key in list(
            self.keys()
        ):  # list() is used to avoid modifying dict during iteration
            if n_params > len(self[key].y):
                warnings.warn(
                    f"Removing key '{key}' from Dataset: number of parameters "
                    f"({n_params}) exceeds number of data points ({len(self[key].y)}).",
                    stacklevel=2,
                )
                del self[key]


# fmt: off
@typing.overload
def binding_1site(
    x: float, K: float, S0: float, S1: float, is_ph: bool = False  # noqa: N803
) -> float: ...


@typing.overload
def binding_1site(
    x: ArrayF, K: float, S0: float, S1: float, is_ph: bool = False  # noqa: N803
) -> ArrayF: ...

# fmt: on
def binding_1site(
    x: float | ArrayF, K: float, S0: float, S1: float, is_ph: bool = False  # noqa: N803
) -> float | ArrayF:  # fmt: skip
    """Single site binding model function.

    Parameters
    ----------
    x : float | ArrayF
        Concentration values.
    K : float
        Dissociation constant.
    S0 : float
        Plateau value for the unbound state.
    S1 : float
        Plateau value for the bound state.
    is_ph : bool, optional
        If True, use the pH model for binding. Default is False.

    Returns
    -------
    float | ArrayF
        Modeled binding values.

    Notes
    -----
    The parameters K, S0 and S1 are in uppercase by convention as used in lmfit library.
    """
    if is_ph:
        return S0 + (S1 - S0) * 10 ** (K - x) / (1 + 10 ** (K - x))
    return S0 + (S1 - S0) * x / K / (1 + x / K)


### helpers
def _binding_1site_models(params: Parameters, x: ArrayDict, is_ph: bool) -> ArrayDict:
    """Compute models for the given input data and parameters."""
    models = {}
    for lbl, x_data in x.items():
        models[lbl] = binding_1site(
            x_data,
            params["K"].value,
            params[f"S0_{lbl}"].value,
            params[f"S1_{lbl}"].value,
            is_ph,
        )
    return models


def _init_from_dataset(ds: Dataset) -> tuple[ArrayDict, ArrayDict, ArrayDict, bool]:
    x = {k: da.x for k, da in ds.items()}
    y = {k: da.y for k, da in ds.items()}
    w = {k: da.w if da.w is not None else np.ones_like(da.y) for k, da in ds.items()}
    return x, y, w, ds.is_ph


def _binding_1site_residuals(params: Parameters, ds: Dataset) -> ArrayF:
    """Compute concatenated residuals (array) for multiple datasets [weight = 1/sigma]."""
    x, y, w, is_ph = _init_from_dataset(ds)
    models = _binding_1site_models(params, x, is_ph)
    residuals: ArrayF = np.concatenate([(w[lbl] * (y[lbl] - models[lbl])) for lbl in x])
    return residuals


def kd(kd1: float, pka: float, ph: ArrayF | float) -> ArrayF | float:
    """Infinite cooperativity model.

    It can describe pH-dependence for chloride dissociation constant.

    Parameters
    ----------
    kd1 : float
        Dissociation constant at pH <= 5.0 (fully protonated).
    pka : float
        Acid dissociation constant.
    ph : ArrayF | float
        pH value(s).

    Returns
    -------
    ArrayF | float
        Predicted Kd value(s).

    Examples
    --------
    >>> kd(10, 8.4, 7.4)
    11.0
    >>> import numpy as np
    >>> kd(10, 8.4, np.array([7.4, 8.4]))
    array([11., 20.])
    """
    return kd1 * (1 + 10 ** (pka - ph)) / 10 ** (pka - ph)


def _build_params_1site(ds: Dataset) -> Parameters:
    """Initialize parameters for 1 site model based on the given dataset."""
    params = Parameters()
    k_initial = []
    for lbl, da in ds.items():
        params.add(f"S0_{lbl}", value=da.y[0])
        params.add(f"S1_{lbl}", value=da.y[-1])
        target_y = (da.y[0] + da.y[-1]) / 2
        k_initial.append(da.x[np.argmin(np.abs(da.y - target_y))])
    if ds.is_ph:
        params.add("K", value=np.mean(k_initial), min=3, max=11)
    else:
        # epsilon avoids x/K raise x/0 error
        params.add("K", value=np.mean(k_initial), min=float_info.epsilon)
    return params


@dataclass
class FitResult:
    """A dataclass representing the results of a fit.

    Attributes
    ----------
    figure : mpl.figure.Figure
        A matplotlib figure object representing the plot of the fit result.
    result : MinimizerResult
        The minimizer result object representing the outcome of the fit.
    mini : Minimizer
        The Minimizer object used for the fit.
    """

    figure: figure.Figure | None = None
    result: MinimizerResult | None = None
    mini: Minimizer | None = None

    def is_valid(self) -> bool:
        """Whether figure, result, and minimizer exist."""
        return (
            self.figure is not None
            and self.result is not None
            and self.mini is not None
        )


@dataclass
class SpectraGlobResults:
    """A dataclass representing the results of both svd and bands fits.

    Attributes
    ----------
    svd : FitResult | None
        The `FitResult` object representing the outcome of the concatenated svd
        fit, or `None` if the svd fit was not performed.
    gsvd : FitResult | None
        The `FitResult` object representing the outcome of the svd fit, or
        `None` if the svd fit was not performed.
    bands : FitResult | None
        The `FitResult` object representing the outcome of the bands fit, or
        `None` if the bands fit was not performed.
    """

    svd: FitResult | None = field(default=None)
    gsvd: FitResult | None = field(default=None)
    bands: FitResult | None = field(default=None)


def analyze_spectra(
    spectra: pd.DataFrame, is_ph: bool, band: tuple[int, int] | None = None
) -> FitResult:
    """Analyze spectra titration, create and plot fit results.

    This function performs either Singular Value Decomposition (SVD) or
    integrates the spectral data over a specified band and fits the integrated
    data to a binding model.

    Parameters
    ----------
    spectra : pd.DataFrame
        The DataFrame containing spectra (one spectrum for each column).
    is_ph : bool
        Whether the x values should be interpreted as pH values rather than
        concentrations.
    band : tuple[int, int] | None
        The band to integrate over. If None (default), performs Singular Value
        Decomposition (SVD).

    Returns
    -------
    FitResult
        A FitResult object containing the figure, result, and mini.

    Raises
    ------
    ValueError
        If the band parameters are not in the spectra's index when the band
        method is used.

    Notes
    -----
    Creates plots of spectra, principal component vectors, singular values, fit
    of the first principal component and PCA for SVD; only of spectra and fit
    for Band method.
    """
    y_offset = 1.0
    x = spectra.columns.to_numpy()
    fig, (ax1, ax2, ax3, ax4, ax5) = _create_spectra_canvas()
    plot_spectra(ax1, spectra, PlotParameters(is_ph))
    if band is None:  # SVD
        threshold = int(len(spectra) * 0.5)
        spectra = spectra.dropna(axis=1, thresh=threshold)
        spectra = spectra.dropna(axis=0)
        ddf = spectra.sub(spectra.iloc[:, 0], axis=0)
        u, s, v = np.linalg.svd(ddf)
        ds = Dataset(x, v[0, :] + y_offset, is_ph)
        plot_autovectors(ax2, spectra.index, u)
        plot_autovalues(ax3, s[:])  # don't plot last auto-values?
        plot_pca(ax5, v, x, PlotParameters(is_ph))
        ylabel = "First Principal Component"
        ylabel_color = COLOR_MAP(0)
    else:  # Band integration
        ini, fin = band
        if ini not in spectra.index and fin not in spectra.index:
            msg = f"Band parameters ({ini}, {fin}) are not in the spectra's index."
            raise ValueError(msg)
        # columns index name are not necessarily unique
        y = np.array(
            [spectra.iloc[:, i].loc[ini:fin].sum() for i in range(spectra.shape[1])]
        )
        # rescale y
        y /= np.abs(y).max() / 10
        ds = Dataset(x, y, is_ph)
        ylabel = "Integrated Band Fluorescence"
        ylabel_color = (0.0, 0.0, 0.0, 1.0)  # "k"
    weight_multi_ds_titration(ds)
    fit_result = fit_binding_glob(ds)
    result = fit_result.result
    plot_fit(ax4, ds, result, nboot=N_BOOT, pp=PlotParameters(is_ph))
    ax4.set_ylabel(ylabel, color=ylabel_color)
    return FitResult(fig, result, fit_result.mini)


class InsufficientDataError(Exception):
    """Raised to prevent fitting failure for too few data points."""


def fit_binding_glob(ds: Dataset) -> FitResult:
    """Analyze multi-label titration datasets and visualize the results."""
    params = _build_params_1site(ds)
    if len(params) > len(np.concatenate([da.y for da in ds.values()])):
        raise InsufficientDataError
    mini = Minimizer(_binding_1site_residuals, params, fcn_args=(ds,))
    result = mini.minimize()
    fig = figure.Figure()
    ax = fig.add_subplot(111)
    plot_fit(ax, ds, result, nboot=N_BOOT, pp=PlotParameters(ds.is_ph))
    return FitResult(fig, result, mini)


# TODO: remove the print statements use logging
def weight_individual_ds(ds: Dataset) -> None:
    """Assign weights to each label based on individual label fitting."""
    if len(ds.keys()) != 1:
        print("Expect single label dataset.")
        return
    params = _build_params_1site(ds)
    da = next(iter(ds.values()))
    if len(params) > len(da.y):
        print("failed")
        w = 1.0 * np.ones_like(da.x)
    else:
        res = lmfit.minimize(_binding_1site_residuals, params, args=(ds,))
        # Calculate residuals SEM
        sem = np.std(res.residual, ddof=1) / np.sqrt(len(res.residual))
        w = 1 / sem * np.ones_like(da.x)
    ds.add_weights(w)


def weight_multi_ds_titration(ds: Dataset) -> None:
    """Assign weights to each label based on individual label fitting."""
    w_d: ArrayDict = {}
    failed_fit_labels = []
    for lbl in ds:
        d = ds.copy(keys=[lbl])
        weight_individual_ds(d)
        w_d[lbl] = next(iter(d.values())).w  # type: ignore[assignment]
        if np.all(w_d[lbl] == 1.0):
            failed_fit_labels.append(lbl)
    if failed_fit_labels:
        minimum_weight = np.inf
        for lbl in ds.keys() - set(failed_fit_labels):
            minimum_weight = min(np.min(w_d[lbl]).item(), minimum_weight)
        for lbl in failed_fit_labels:
            w_d[lbl] *= minimum_weight / 10
    ds.add_weights(w_d)


def analyze_spectra_glob(
    titration: dict[str, pd.DataFrame],
    ds: Dataset,
    dbands: dict[str, tuple[int, int]] | None = None,
) -> SpectraGlobResults:
    """Analyze multi-label spectra visualize the results."""
    _gap_ = 1
    dbands = dbands or {}
    labels_svd = titration.keys() - dbands.keys()
    labels_bands = titration.keys() - labels_svd
    if len(labels_svd) > 1:
        # Concatenate spectra.
        prev_max = 0
        adjusted_list = []
        for lbl in labels_svd:
            spectra_adjusted = titration[lbl].copy()  # Avoid modifying original data
            spectra_adjusted.index += prev_max - spectra_adjusted.index.min() + _gap_
            prev_max = spectra_adjusted.index.max()
            adjusted_list.append(spectra_adjusted)
        spectra_merged = pd.concat(adjusted_list)
        svd = analyze_spectra(spectra_merged, ds.is_ph)
        ds_svd = ds.copy(labels_svd)
        weight_multi_ds_titration(ds)
        f_res = fit_binding_glob(ds_svd)
        fig = _plot_spectra_glob_emcee(titration, ds_svd, f_res)
        gsvd = FitResult(fig, f_res.result, f_res.mini)
    else:
        svd, gsvd = None, None
    if len(labels_bands) > 1:
        ds_bands = ds.copy(labels_bands)
        weight_multi_ds_titration(ds)
        f_res = fit_binding_glob(ds_bands)
        fig = _plot_spectra_glob_emcee(titration, ds_bands, f_res, dbands)
        bands = FitResult(fig, f_res.result, f_res.mini)
    else:
        bands = None
    return SpectraGlobResults(svd, gsvd, bands)


def plot_fit(
    ax: axes.Axes,
    ds: Dataset,
    result: MinimizerResult,
    nboot: int = 0,
    pp: PlotParameters | None = None,
) -> None:
    """Plot residuals for each dataset with uncertainty."""
    _stretch = 0.05
    xfit = {
        k: np.linspace(da.x.min() * (1 - _stretch), da.x.max() * (1 + _stretch), 100)
        for k, da in ds.items()
    }
    yfit = _binding_1site_models(result.params, xfit, ds.is_ph)
    # Create a color cycle
    colors = [COLOR_MAP(i) for i in range(len(ds))]
    for (lbl, da), clr in zip(ds.items(), colors, strict=False):
        # Make sure a label will be displayed.
        label = lbl if (da.w is None and nboot == 0) else None
        # Plot data.
        if pp:
            ax.scatter(
                da.x,
                da.y,
                c=list(da.x),
                s=99,
                edgecolors=clr,
                label=label,
                vmin=pp.hue_norm[0],
                vmax=pp.hue_norm[1],
                cmap=pp.palette,
            )
        else:
            ax.plot(da.x, da.y, "o", color=clr, label=label)
        # Plot fitting.
        ax.plot(xfit[lbl], yfit[lbl], "-", color="gray")
        if nboot:
            # Calculate uncertainty using Monte Carlo method.
            y_samples = np.empty((nboot, len(xfit[lbl])))
            rng = np.random.default_rng()
            for i in range(nboot):
                p_sample = result.params.copy()
                for param in p_sample.values():
                    # Especially stderr can be None in case of critical fitting
                    if param.value and param.stderr:
                        param.value = rng.normal(param.value, param.stderr)
                y_samples[i, :] = _binding_1site_models(p_sample, xfit, ds.is_ph)[lbl]
            dy = y_samples.std(axis=0)
            # Plot uncertainty.
            # Display label in fill_between plot.
            ax.fill_between(
                xfit[lbl],
                yfit[lbl] - dy,
                yfit[lbl] + dy,
                alpha=0.4,
                color=clr,
                label=lbl,
            )
            if da.w is not None:
                ye = 1 / da.w
                ax.errorbar(
                    da.x, da.y, yerr=ye, fmt="none", color="gray", alpha=0.4, capsize=3
                )
        elif da.w is not None:
            ye = 1 / da.w
            # Display label in error bar plot.
            ax.errorbar(
                da.x, da.y, yerr=ye, fmt=".", label=lbl, color=clr, alpha=0.4, capsize=3
            )
    ax.legend()
    if result.params["K"].stderr:  # Can be None in case of critical fitting
        k = ufloat(result.params["K"].value, result.params["K"].stderr)
    else:
        k = f'{result.params["K"].value:.3g}' if result.params["K"].value else None
    title = "=".join(["K", str(k).replace("+/-", "±")])
    xlabel = "pH" if ds.is_ph else "Cl"
    _apply_common_plot_style(ax, f"LM fit {title}", xlabel, "")


def plot_fit2(
    ds: Dataset,
    xerr: ArrayF,
    params: Parameters,
    nboot: int = 0,
    pp: PlotParameters | None = None,
) -> figure.Figure:
    """Plot residuals for each dataset with uncertainty."""
    _stretch = 0.05
    xfit = {
        k: np.linspace(da.x.min() * (1 - _stretch), da.x.max() * (1 + _stretch), 100)
        for k, da in ds.items()
    }
    yfit = _binding_1site_models(params, xfit, ds.is_ph)
    fig, ax = plt.subplots(1, 1, figsize=(7, 7))
    # Create a color cycle
    colors = [COLOR_MAP(i) for i in range(len(ds))]
    ylim = (np.inf, -np.inf)
    for (lbl, da), clr in zip(ds.items(), colors, strict=False):
        # Plot data.
        if pp:
            ax.scatter(
                da.x,
                da.y,
                c=list(da.x),
                s=69,
                edgecolors=clr,
                label=lbl,
                vmin=pp.hue_norm[0],
                vmax=pp.hue_norm[1],
                cmap=pp.palette,
            )
        else:
            ax.plot(da.x, da.y, "o", color=clr, label=lbl)
        # Plot fitting.
        ax.plot(xfit[lbl], yfit[lbl], "-", color="gray")
        if da.w is not None:
            ye = 1 / da.w
            ax.errorbar(
                da.x,
                da.y,
                yerr=ye,
                xerr=xerr,
                fmt="none",
                color="k",
                alpha=0.2,
                capsize=3,
            )
        current_ylim = ax.get_ylim()
        ylim = (min(current_ylim[0], ylim[0]), max(current_ylim[1], ylim[1]))
        if nboot:
            # Calculate uncertainty using Monte Carlo method.
            y_samples = np.empty((nboot, len(xfit[lbl])))
            rng = np.random.default_rng()
            for i in range(nboot):
                p_sample = params.copy()
                for param in p_sample.values():
                    # Especially stderr can be None in case of critical fitting
                    if param.value and param.stderr:
                        param.value = rng.normal(param.value, param.stderr)
                y_samples[i, :] = _binding_1site_models(p_sample, xfit, ds.is_ph)[lbl]
            dy = y_samples.std(axis=0)
            # Plot uncertainty.
            # Display label in fill_between plot.
            ax.fill_between(
                xfit[lbl], yfit[lbl] - dy, yfit[lbl] + dy, alpha=0.4, color=clr
            )
    ax.set_ylim(ylim)
    ax.legend()
    if params["K"].stderr:  # Can be None in case of critical fitting
        k = ufloat(params["K"].value, params["K"].stderr)
    else:
        k = f'{params["K"].value:.3g}' if params["K"].value else None
    title = "=".join(["K", str(k).replace("+/-", "±")])
    xlabel = "pH" if ds.is_ph else "Cl"
    _apply_common_plot_style(ax, f"LM fit {title}", xlabel, "")
    plt.close()
    return fig


def _plot_spectra_glob_emcee(
    titration: dict[str, pd.DataFrame],
    ds: Dataset,
    f_res: FitResult,
    dbands: dict[str, tuple[int, int]] | None = None,
) -> figure.Figure:
    fig, (ax1, ax2, ax3, ax4, ax5) = _create_spectra_canvas()
    fig.delaxes(ax1)
    fig.delaxes(ax2)
    fig.delaxes(ax3)
    pparams = PlotParameters(ds.is_ph)
    tit_filtered = {k: spec for k, spec in titration.items() if k in ds}
    plot_spectra_distributed(fig, tit_filtered, pparams, dbands)
    plot_fit(ax4, ds, f_res.result, nboot=N_BOOT, pp=pparams)
    if f_res.mini:
        result_emcee = f_res.mini.emcee(
            steps=EMCEE_STEPS * 3, workers=8, burn=100, nwalkers=30, progress=False
        )
        plot_emcee_k_on_ax(ax5, result_emcee)
    return fig


# ODR ########################################################################
def format_estimate(
    value: float, error: float, significant_digit_limit: int = 5
) -> str:
    """Format the value and its associated error into "{value} ± {error}" string."""
    large_number = np.exp(significant_digit_limit)
    small_number = np.exp(-significant_digit_limit)
    significant_digits = max(0, -int(np.floor(np.log10(abs(error)))))
    use_scientific = (
        significant_digits > significant_digit_limit
        or abs(value) > large_number
        or abs(value) < small_number
    )
    formatted_value = (
        f"{value:.{significant_digits + 1}e}"
        if use_scientific
        else f"{value:.{significant_digits + 1}f}"
    )
    formatted_error = (
        f"{error:.{significant_digits + 1}e}"
        if use_scientific
        else f"{error:.{significant_digits + 1}f}"
    )
    return f"{formatted_value} ± {formatted_error}"


@dataclass
class _ODRResult:
    parameters: Parameters
    output: odr.Output


def odr_fitting(
    x: ArrayF,
    x_err: ArrayF,
    y_list: list[ArrayF],
    y_err_list: list[ArrayF],
    pars: Parameters,
) -> _ODRResult:
    """Fit ODR model on two datasets."""
    # Combine y1 and y2 data and errors
    n_dataset = len(y_list)
    y_data = np.concatenate(y_list)
    y_err = np.concatenate(y_err_list)
    x_data_combined = np.tile(x, n_dataset)
    x_err_combined = np.tile(x_err, n_dataset)

    data = odr.RealData(x_data_combined, y_data, sx=x_err_combined, sy=y_err)

    def model_function(pars: list[float], x: ArrayF) -> ArrayF:
        K, S0, S1 = pars  # noqa: N806
        return binding_1site(x, K, S0, S1, is_ph=True)

    def combined_model(pars: list[float], x: ArrayF) -> ArrayF:
        K, S0_1, S1_1, S0_2, S1_2 = pars  # noqa: N806
        y1_model = model_function([K, S0_1, S1_1], x[: len(x) // 2])
        y2_model = model_function([K, S0_2, S1_2], x[len(x) // 2 :])
        return np.concatenate((y1_model, y2_model))

    combined_model_odr = odr.Model(combined_model)
    # combined_model_odr = odr.Model(model_function)  # noqa: ERA001

    keys = ["K", "S0_y1", "S1_y1", "S0_y2", "S1_y2"]
    # keys = ["K", "S0_default", "S1_default"]  # noqa: ERA001
    initial_params = [pars[key].value for key in keys]

    # Create ODR objects
    odr_obj = odr.ODR(data, combined_model_odr, beta0=initial_params)
    # Run the fitting
    output = odr_obj.run()

    params = Parameters()
    for name, value, error in zip(keys, output.beta, output.sd_beta, strict=True):
        params.add(name, value=value)
        params[name].stderr = error
    return _ODRResult(params, output)



"""Command-line interface."""

from __future__ import annotations

import csv
import pprint
from collections import namedtuple
from pathlib import Path
from typing import Any

import click
import lmfit  # type: ignore[import-untyped]
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from click import Context, Path as cPath

from clophfit import __enspire_out_dir__, __tecan_out_dir__, binding, prenspire, prtecan
from clophfit.prenspire import EnspireFile
from clophfit.prtecan import TecanConfig, Titration


@click.group()
@click.version_option(message="%(version)s")
def clop() -> None:  # pragma: no cover
    """Group command."""


@clop.command()
@click.argument("kd1", type=float)
@click.argument("pka", type=float)
@click.argument("ph", type=float)
def eq1(kd1: float, pka: float, ph: float) -> None:
    """Model Kd dependence on pH."""
    click.echo(binding.fitting.kd(kd1=kd1, pka=pka, ph=ph))


@click.group()
@click.pass_context
@click.version_option(message="%(version)s")
@click.option("--verbose", "-v", count=True, help="Verbosity of messages.")
@click.option("--out", "-o", type=cPath(), help="Output folder.")
def ppr(ctx: Context, verbose: int, out: str) -> None:  # pragma: no cover
    """Parse Plate Reader `ppr` group command."""
    ctx.ensure_object(dict)
    ctx.obj["VERBOSE"] = verbose
    if out:
        ctx.obj["OUT"] = out


######################################
# pr.tecan                           #
######################################
@ppr.command()
@click.pass_context
@click.argument("list_file", type=cPath(exists=True))
@click.option("--cl", type=float, help="Cl titration: [Cl] (mM) of added aliquots.")
@click.option("--bg", is_flag=True, help="Subtract buffer (from scheme.txt).")
@click.option("--bg-adj", is_flag=True, help="Adjust bg to avoid negative value.")
@click.option("--dil", is_flag=True, help="Apply dilution correction.")
@click.option("--nrm", is_flag=True, help="Normalize using metadata.")
@click.option("--bg-mth", default="mean", show_default=True, help="Method for bg.")
@click.option("--sch", type=cPath(exists=True), help="Plate scheme (buffers CTRs).")
@click.option("--add", type=cPath(exists=True), help="Initial volume and additions.")
@click.option(
    "--all", "comb", is_flag=True, help="Export (fit) all corrections combinations."
)
@click.option("--lim", type=(float, float), help="Range MIN, MAX of plot_K.")
@click.option("--sel", type=(float, float), help="Select from K_MIN S1_MIN.")
@click.option("--title", "-t", type=str, default="", help="Title for plots.")
@click.option("--fit/--no-fit", default=True, show_default=True, help="Perform also fit.")  # fmt: skip
@click.option("--png/--no-png", default=True, show_default=True, help="Export png files.")  # fmt: skip
@click.option("--pdf/--no-pdf", default=False, show_default=True, help="Full report in pdf.")  # fmt: skip
def tecan(  # noqa: PLR0913
    ctx: Context,
    list_file: str,
    cl: float,
    bg: bool,
    bg_adj: bool,
    dil: bool,
    nrm: bool,
    bg_mth: str,
    sch: str | None,
    add: str | None,
    comb: bool,
    lim: tuple[float, float] | None,
    sel: tuple[float, float] | None,
    title: str,
    fit: bool,
    png: bool,
    pdf: bool,
) -> None:
    """Convert a list of Tecan-exported excel files into titrations.

    LIST_FILE : List of Tecan files and concentration values.

    Saves titrations as .dat files and fits all wells using 2 labels. The
    function produces:

    - K plot

    - ebar and (for selection) ebarZ plot

    - all_wells pdf

    - csv tables for all labelblocks and global fittings.

    Note: Buffer is always subtracted if scheme indicates buffer well positions.
    """
    verbose: int = ctx.obj.get("VERBOSE", 0)
    out = ctx.obj.get("OUT", __tecan_out_dir__)
    out_fp = Path(out) / "Cl" if cl else Path(out) / "pH"
    out_fp.mkdir(parents=True, exist_ok=True)
    # Options validation.
    if cl and not add:
        msg = "--cl requires --add to be specified."
        raise click.UsageError(msg)
    if dil and not add:
        msg = "--dil requires --add to be specified."
        raise click.UsageError(msg)
    if bg and not sch:
        # Also --sch must contain valid buffers!
        msg = "Scheme is needed to compute buffer bg i.e. --bg requires --sch!"
        raise click.UsageError(msg)
    if comb and not (bg and sch and dil):
        msg = "All combinations requires --bg and --dil to be specified."
        raise click.UsageError(msg)
    # Config
    tecan_config = TecanConfig(out_fp, verbose, comb, lim, sel, title, fit, png, pdf)
    # Load titration
    list_fp = Path(list_file)
    click.secho(f"** File: {list_fp.resolve()}", fg="green")
    click.echo(tecan_config)
    tit = Titration.fromlistfile(list_fp, not cl)
    tit.params.bg = bg
    tit.params.bg_adj = bg_adj
    tit.params.dil = dil
    tit.params.nrm = nrm
    tit.params.bg_mth = bg_mth
    click.echo(tit.params)
    if add:
        tit.load_additions(Path(add))
    if cl and tit.additions:
        tit.conc = prtecan.calculate_conc(tit.additions, cl)
    if sch:
        tit.load_scheme(Path(sch))
        f = tit.buffer.plot(title=title)
        f.savefig(out_fp / "buffer.png")
        f = tit.buffer.plot(nrm=True, title=title)
        f.savefig(out_fp / "buffer_norm.png")
    with (out_fp / "metadata-labels.txt").open("w", encoding="utf-8") as fp:
        for lbg in tit.labelblocksgroups:
            pprint.pprint(lbg.metadata, stream=fp)
    f = tit.plot_temperature(title=title)
    f.savefig(out_fp / "temperatures.png")
    # Output and export
    tit.export_data_fit(tecan_config)


########################################
# pr.enspire                           #
########################################
@ppr.command()
@click.pass_context
@click.argument("csv_f", type=cPath(exists=True, path_type=str))
@click.argument("note_f", type=cPath(exists=True), required=False)
@click.option(
    "-b",
    "bands",
    multiple=True,
    default=None,
    nargs=3,
    type=(str, int, int),
    help="Label and band interval (format: LABEL LOWER UPPER)",
)
def enspire(ctx: Context, csv_f: str, note_f: str | None, bands: tuple[Any]) -> None:
    """Save spectra as csv tables from EnSpire xls file."""
    verbose = ctx.obj.get("VERBOSE", 0)
    out = ctx.obj.get("OUT", __enspire_out_dir__)
    ef = EnspireFile(Path(csv_f), verbose=verbose)
    ef.export_measurements(Path(out))
    if note_f is not None:
        fit_enspire(ef, Path(note_f), Path(out), list(bands), verbose)


# TODO: Simplify this function
def fit_enspire(  # noqa: C901,PLR0912
    ef: EnspireFile,
    note_fp: Path,
    out_dir: Path,
    bands: list[tuple[str, int, int]],
    verbose: int,
) -> None:
    """Fit prenspire titration (all labels, temp, mutant, titrations)."""
    note = prenspire.prenspire.Note(note_fp, verbose=verbose)
    note.build_titrations(ef)
    dbands = {label: (ini, fin) for label, ini, fin in bands} if bands else {}
    x_combined = {}
    y_combined = {}
    for name, d_name in note.titrations.items():
        for temp, d_temp in d_name.items():
            for tit, d_tit in d_temp.items():
                if tit.split("_")[0] == "pH":
                    is_ph = False
                elif tit.split("_")[0] == "Cl":
                    is_ph = True
                else:
                    msg = "Unknown titration type."
                    raise ValueError(msg)
                for label, data in d_tit.items():
                    band = dbands.get(label)
                    fit_result = binding.fitting.analyze_spectra(data, is_ph, band)
                    if fit_result.is_valid() and fit_result.mini:
                        x_combined[label] = fit_result.mini.userargs[0]["default"].x
                        y_combined[label] = fit_result.mini.userargs[0]["default"].y
                        pdf_file = out_dir / f"{name}_{temp}_{label}_{tit}_{band}.pdf"
                        if fit_result.figure:
                            fit_result.figure.savefig(pdf_file)
                    _print_result(fit_result, pdf_file, str(band))
                # Global spectra analysis with more than 1 label.
                if (
                    len(d_tit.keys() - dbands.keys()) > 1  # svd > 1
                    or len(dbands.keys() & d_tit.keys()) > 1  # bands > 1
                ):
                    ds = binding.fitting.Dataset(x_combined, y_combined, is_ph)
                    spectra_gres = binding.fitting.analyze_spectra_glob(
                        d_tit, ds, dbands
                    )
                    if spectra_gres.svd and spectra_gres.svd.is_valid():
                        pdf_file = out_dir / f"{name}_{temp}_all_{tit}_SVD.pdf"
                        if spectra_gres.svd.figure:
                            spectra_gres.svd.figure.savefig(pdf_file)
                        _print_result(spectra_gres.svd, pdf_file, "")
                    if spectra_gres.gsvd and spectra_gres.gsvd.is_valid():
                        pdf_file = out_dir / f"{name}_{temp}_g_{tit}_SVD.pdf"
                        if spectra_gres.gsvd.figure:
                            spectra_gres.gsvd.figure.savefig(pdf_file)
                        _print_result(spectra_gres.gsvd, pdf_file, "")
                    if spectra_gres.bands and spectra_gres.bands.is_valid():
                        keys = dbands.keys() & d_tit.keys()
                        lname = [f"{k}({dbands[k][0]},{dbands[k][1]})" for k in keys]
                        bands_str = "".join(lname)
                        pdf_file = out_dir / f"{name}_{temp}_all_{tit}_{bands_str}.pdf"
                        if spectra_gres.bands.figure:
                            spectra_gres.bands.figure.savefig(pdf_file)
                        _print_result(spectra_gres.bands, pdf_file, bands_str)


def _print_result(
    fit_result: binding.fitting.FitResult,
    pdf_file: Path,
    band_str: str,
) -> None:
    print(str(pdf_file))
    print(f"Best fit using '{band_str}' band:\n")
    ci = lmfit.conf_interval(fit_result.mini, fit_result.result)
    print(lmfit.ci_report(ci, ndigits=2, with_offset=False))
    print(f"\n Plot saved in '{pdf_file}'.\n")


#############################
#  fit_titration_global     #
#############################
@click.group()
@click.pass_context
@click.version_option(message="%(version)s")
@click.option("--verbose", "-v", count=True, help="Verbosity of messages.")
@click.option("--out", "-o", type=cPath(), help="Output folder.")
@click.option(
    "--is-ph/--no-is-ph", default=True, show_default=True, help="Concentrations are pH."
)
def fit_titration(
    ctx: Context, verbose: int, out: str, is_ph: bool
) -> None:  # pragma: no cover
    """Fit Titration group command."""
    ctx.ensure_object(dict)
    ctx.obj["VERBOSE"] = verbose
    ctx.obj["OUT"] = out
    ctx.obj["IS_PH"] = is_ph


@fit_titration.command()
@click.pass_context
@click.argument("csv_f", type=click.Path(exists=True))
@click.argument("note_f", type=click.Path(exists=True))
@click.option(
    "-b", "--band", nargs=2, type=int, help="Integration interval from <1> to <2>"
)
def spec(ctx: Context, csv_f: str, note_f: str, band: tuple[int, int] | None) -> None:
    """Update old svd or band fit of titration spectra."""
    verbose = ctx.obj.get("VERBOSE", 0)
    is_ph = ctx.obj.get("IS_PH", True)
    out = Path(ctx.obj.get("OUT", "."))

    note_df = pd.read_csv(note_f, sep="\t")
    csv_df = pd.read_csv(csv_f)
    # Ignore buffer wells! SVD will use differences between spectra.
    note_df = note_df[note_df["mutant"] != "buffer"]
    Notes = namedtuple("Notes", ["wells", "conc"])
    titration_type = "pH" if is_ph else "Cl"
    note = Notes(list(note_df["well"]), list(note_df[titration_type]))
    spectra = csv_df[note.wells]
    spectra.index = csv_df["lambda"]
    spectra.columns = np.array(note.conc)
    if verbose:
        print(csv_df)
        click.echo(note_f)
        print(note)
        print("DataFrame\n", spectra)
    is_ph = titration_type == "pH"
    fit_result = binding.fitting.analyze_spectra(spectra, is_ph, band)
    # output
    out.mkdir(parents=True, exist_ok=True)
    pdf_file = out / f"{Path(csv_f).stem}_{band}_{Path(note_f).stem}.pdf"
    if fit_result.figure is not None:
        fit_result.figure.savefig(pdf_file)
    _print_result(fit_result, pdf_file, str(band))


@fit_titration.command()
@click.pass_context
@click.argument("file", type=click.Path(exists=True))
@click.option("-b", "--boot", type=int, help="Number of booting iterations.")
@click.option(
    "--weight/--no-weight", default=True, show_default=True, help="Use residue weights."
)
def glob(ctx: Context, file: str, boot: int, weight: bool) -> None:
    """Update old glob fit of multiple datasets."""
    verbose = ctx.obj.get("VERBOSE", 0)
    is_ph = ctx.obj.get("IS_PH", True)
    file_df = pd.read_csv(file)
    fp = Path(file)
    min_correl_to_print = 0.65
    burn = 75
    if verbose:
        click.echo(file_df)
    yc = {lbl: file_df[lbl].to_numpy() for lbl in file_df.columns[1:]}
    ds = binding.fitting.Dataset(file_df["x"].to_numpy(), yc, is_ph)
    if weight:
        binding.fitting.weight_multi_ds_titration(ds)
    f_res = binding.fitting.fit_binding_glob(ds)
    # Figure
    figure, ax = plt.subplots()
    binding.fitting.plot_fit(
        ax, ds, f_res.result, nboot=30, pp=binding.plotting.PlotParameters(is_ph)
    )
    lmfit.printfuncs.report_fit(f_res.result, min_correl=min_correl_to_print)
    figure.savefig(Path(file).with_suffix(".png"))
    if boot and f_res.mini:
        # Emcee
        samples = f_res.mini.emcee(burn=burn, steps=boot).flatchain
        fig = binding.plotting.plot_emcee(samples)
        fig.savefig(fp.with_suffix(".png").with_stem(fp.stem + "-emcee"))
        hdi = samples.quantile([0.025, 0.975])["K"].to_list()
        print(f"Quantiles for K: {[f'{q:.3g}' for q in hdi]}")
        hdi = samples.quantile([0.03, 0.97])["K"].to_list()
        print(f"HDI (94%): {[f'{q:.3g}' for q in hdi]}")
        # R := S0 / S1
        # function := S1 * (R + (1 - R) * 10 ** (K - x) / (1 + 10 ** (K - x)))
        if is_ph:  # ratio between protonated un-protonated states
            ratios = {lbl: samples[f"S0_{lbl}"] / samples[f"S1_{lbl}"] for lbl in ds}
            # Combine ratio and K samples into a DataFrame for corner plot
            samples_ratios = pd.DataFrame({**ratios, **{"K": samples["K"]}})
            fig_ratio = binding.plotting.plot_emcee(samples_ratios)
            fig_ratio.savefig(fp.with_suffix(".png").with_stem(fp.stem + "-emc-ratios"))
            for lbl in ds:
                hdi = samples_ratios.quantile([0.025, 0.5, 0.975])[lbl].to_list()
                formatted_hdi = [f"{q:.3g}" for q in hdi]
                print(f"HDI (94%) for plateau ratio in dataset {lbl}: {formatted_hdi}")


########################################
# note_to_csv                          #
########################################
@click.command()
@click.argument("note", type=click.Path(exists=True))
@click.option("-o", "--output", default=None, help="Output CSV file.")
@click.option("-l", "--labels", default="A B", help="Labels to be appended.")
@click.option("-t", "--temp", default="37.0", help="Temperature to be appended.")
def note2csv(note: str, output: str, labels: str, temp: str) -> None:
    """Convert a tab-separated data file into a CSV file."""
    headers = ["Well", "pH", "Cl", "Name", "Temp", "Labels"]

    input_path = Path(note)
    output_path = Path(output) if output else input_path.with_suffix(".csv")

    if not output_path.exists():
        output_path.write_text(",".join(headers) + "\n", encoding="utf-8")

    # read data from note file, append to output file
    with input_path.open("r", encoding="utf-8") as datafile:
        reader = csv.reader(datafile, delimiter="\t")  # assuming tab-separated values
        next(reader)  # skip the header row

        with output_path.open("a", encoding="utf-8") as f:
            writer = csv.writer(f, lineterminator="\n")
            for row in reader:
                new_row = row[:4] + [temp, labels]
                writer.writerow(new_row)
